# methods in cngi_io.image.image should be used going forward,
# methods in cngi_io.image.cngi_image_io are deprecated
from .image import (
    load_image,
    make_empty_aperture_image,
    make_empty_lmuv_image,
    make_empty_sky_image,
    read_image,
    write_image,
)

__all__ = [
    "load_image",
    "make_empty_aperture_image",
    "make_empty_lmuv_image",
    "make_empty_sky_image",
    "read_image",
    "write_image",
]



from . import schema_table


def setup(app):
    app.add_directive(
        "xradio_array_schema_table", schema_table.ArraySchemaTableDirective
    )
    app.add_directive(
        "xradio_dataset_schema_table", schema_table.DatasetSchemaTableDirective
    )
    app.add_directive("xradio_dict_schema_table", schema_table.DictSchemaTableDirective)
    return {"version": "0.1.0"}



import copy
import xarray as xr
import zarr
import s3fs
import os
from botocore.exceptions import NoCredentialsError

# from xradio.vis._vis_utils._ms.msv2_to_msv4_meta import (
#     column_description_casacore_to_msv4_measure,
# )


def _get_file_system_and_items(ps_store: str):

    # default to assuming the data are accessible on local file system
    if os.path.isdir(ps_store):
        # handle a common shell convention
        if ps_store.startswith("~"):
            ps_store = os.path.expanduser(ps_store)
        items = os.listdir(ps_store)
        file_system = os

    elif ps_store.startswith("s3"):
        # only if not found locally, check if dealing with an S3 bucket URL
        # if not ps_store.endswith("/"):
        #     # just for consistency, as there is no os.path equivalent in s3fs
        #     ps_store = ps_store + "/"

        try:
            # initialize the S3 "file system", first attempting to use pre-configured credentials
            file_system = s3fs.S3FileSystem(anon=False, requester_pays=False)
            items = [
                bd.split(sep="/")[-1]
                for bd in file_system.listdir(ps_store, detail=False)
            ]

        except (NoCredentialsError, PermissionError) as e:
            # only public, read-only buckets will be accessible
            # we will want to add messaging and error handling here
            file_system = s3fs.S3FileSystem(anon=True)
            items = [
                bd.split(sep="/")[-1]
                for bd in file_system.listdir(ps_store, detail=False)
            ]
    else:
        raise FileNotFoundError(
            f"Could not find {ps_store} either locally or in the cloud."
        )

    items = [
        item for item in items if not item.startswith(".")
    ]  # Mac OS likes to place hidden files in the directory (.DStore).
    return file_system, items


def _open_dataset(
    store, file_system=os, xds_isel=None, data_variables=None, load=False
):
    """

    Parameters
    ----------
    store : _type_
        _description_
    xds_isel : _type_, optional
        Example {'time':slice(0,10), 'frequency':slice(5,7)}, by default None
    data_variables : _type_, optional
        Example ['VISIBILITY','WEIGHT'], by default None
    load : bool, optional
        _description_, by default False

    Returns
    -------
    _type_
        _description_
    """

    import dask

    if isinstance(file_system, s3fs.core.S3FileSystem):
        mapping = s3fs.S3Map(root=store, s3=file_system, check=False)
        xds = xr.open_zarr(store=mapping)
    else:
        xds = xr.open_zarr(store)

    if xds_isel is not None:
        xds = xds.isel(xds_isel)

    if data_variables is not None:
        xds_sub = xr.Dataset()
        for dv in data_variables:
            xds_sub[dv] = xds[dv]
        xds_sub.attrs = xds.attrs
        xds = xds_sub

    if load:
        with dask.config.set(scheduler="synchronous"):
            xds = xds.load()
    return xds


# Code to depricate:
def _get_attrs(zarr_obj):
    """
    get attributes of zarr obj (groups or arrays)
    """
    return {k: v for k, v in zarr_obj.attrs.asdict().items() if not k.startswith("_NC")}


def _load_no_dask_zarr(zarr_name, slice_dict={}):
    """
    Alternative to xarray open_zarr where the arrays are not Dask Arrays.

    slice_dict: A dictionary of slice objects for which values to read form a dimension.
                For example silce_dict={'time':slice(0,10)} would select the first 10 elements in the time dimension.
                If a dim is not specified all values are retruned.
    return:
        xarray.Dataset()

    #Should go into general utils.
    """
    # Used by xarray to store array labeling info in zarr meta data.
    DIMENSION_KEY = "_ARRAY_DIMENSIONS"

    # logger = _get_logger()
    zarr_group = zarr.open_group(store=zarr_name, mode="r")
    group_attrs = _get_attrs(zarr_group)

    slice_dict_complete = copy.deepcopy(slice_dict)
    coords = {}
    xds = xr.Dataset()
    for var_name, var in zarr_group.arrays():
        print("Hallo 3", var_name, var.shape)
        var_attrs = _get_attrs(var)

        for dim in var_attrs[DIMENSION_KEY]:
            if dim not in slice_dict_complete:
                slice_dict_complete[dim] = slice(None)  # No slicing.

        if (var_attrs[DIMENSION_KEY][0] == var_name) and (
            len(var_attrs[DIMENSION_KEY]) == 1
        ):
            coord = var[
                slice_dict_complete[var_attrs[DIMENSION_KEY][0]]
            ]  # Dimension coordinates.
            del var_attrs["_ARRAY_DIMENSIONS"]
            xds = xds.assign_coords({var_name: coord})
            xds[var_name].attrs = var_attrs
        else:
            # Construct slicing
            slicing_list = []
            for dim in var_attrs[DIMENSION_KEY]:
                slicing_list.append(slice_dict_complete[dim])
            slicing_tuple = tuple(slicing_list)

            print(var_attrs[DIMENSION_KEY])

            xds[var_name] = xr.DataArray(
                var[slicing_tuple], dims=var_attrs[DIMENSION_KEY]
            )

            if "coordinates" in var_attrs:
                del var_attrs["coordinates"]
            del var_attrs["_ARRAY_DIMENSIONS"]
            xds[var_name].attrs = var_attrs

    xds.attrs = group_attrs

    return xds



import toolviper.utils.logger as logger
import time
from typing import Tuple, Union

import numpy as np
import xarray as xr
import os

from xradio.measurement_set._utils._msv2.subtables import subt_rename_ids
from xradio.measurement_set._utils._msv2._tables.read import (
    load_generic_table,
    convert_casacore_time,
    convert_casacore_time_to_mjd,
    make_taql_where_between_min_max,
    table_exists,
)
from xradio._utils.schema import convert_generic_xds_to_xradio_schema
from xradio.measurement_set._utils._msv2.msv4_sub_xdss import (
    rename_and_interpolate_to_time,
)

from xradio._utils.list_and_array import (
    check_if_consistent,
    unique_1d,
    to_list,
    to_np_array,
)


def create_antenna_xds(
    in_file: str,
    spectral_window_id: int,
    antenna_id: list,
    feed_id: list,
    telescope_name: str,
    partition_polarization: xr.DataArray,
) -> xr.Dataset:
    """
    Create an Xarray Dataset containing antenna information.

    Parameters
    ----------
    in_file : str
        Path to the input MSv2.
    spectral_window_id : int
        Spectral window ID.
    antenna_id : list
        List of antenna IDs.
    feed_id : list
        List of feed IDs.
    telescope_name : str
        Name of the telescope.
    partition_polarization: xr.DataArray
        Polarization labels of this partition, needed if that info is not present in FEED

    Returns
    ----------
        xr.Dataset: Xarray Dataset containing the antenna information.
    """
    ant_xds = xr.Dataset(attrs={"type": "antenna"})

    ant_xds = extract_antenna_info(ant_xds, in_file, antenna_id, telescope_name)

    ant_xds = extract_feed_info(
        ant_xds, in_file, antenna_id, feed_id, spectral_window_id
    )
    # Needed for special SPWs such as ALMA WVR or CHANNEL_AVERAGE data (have no feed info)
    if "polarization_type" not in ant_xds:
        pols_chars = list(partition_polarization.values[0])
        pols_labels = [f"pol_{idx}" for idx in np.arange(0, len(pols_chars))]
        ant_xds = ant_xds.assign_coords(receptor_label=pols_labels)
        pol_type_values = [pols_chars] * len(ant_xds.antenna_name)
        ant_xds = ant_xds.assign_coords(
            polarization_type=(
                ["antenna_name", "receptor_label"],
                pol_type_values,
            )
        )

    ant_xds.attrs["overall_telescope_name"] = telescope_name
    return ant_xds


def extract_antenna_info(
    ant_xds: xr.Dataset, in_file: str, antenna_id: list, telescope_name: str
) -> xr.Dataset:
    """Reformats MSv2 Antenna table content to MSv4 schema.

    Parameters
    ----------
    ant_xds : xr.Dataset
        The dataset that will be updated with antenna information.
    in_file : str
        Path to the input MSv2.
    antenna_id : list
        A list of antenna IDs to extract information for.
    telescope_name : str
        The name of the telescope.

    Returns
    -------
    xr.Dataset
        Dataset updated to contain the antenna information.
    """
    to_new_data_variables = {
        "POSITION": ["ANTENNA_POSITION", ["antenna_name", "cartesian_pos_label"]],
        "DISH_DIAMETER": ["ANTENNA_DISH_DIAMETER", ["antenna_name"]],
    }

    to_new_coords = {
        "NAME": ["antenna_name", ["antenna_name"]],
        "STATION": ["station", ["antenna_name"]],
        "MOUNT": ["mount", ["antenna_name"]],
        # "PHASED_ARRAY_ID": ["phased_array_id", ["antenna_name"]],
        "antenna_id": ["antenna_id", ["antenna_name"]],
    }

    # Read ANTENNA table into a Xarray Dataset.
    unique_antenna_id = unique_1d(
        antenna_id
    )  # Also ensures that it is sorted otherwise TaQL will give wrong results.

    generic_ant_xds = load_generic_table(
        in_file,
        "ANTENNA",
        rename_ids=subt_rename_ids["ANTENNA"],
        taql_where=f" where (ROWID() IN [{','.join(map(str,unique_antenna_id))}])",  # order is not guaranteed
    )
    generic_ant_xds = generic_ant_xds.assign_coords({"antenna_id": unique_antenna_id})
    generic_ant_xds = generic_ant_xds.sel(
        antenna_id=antenna_id, drop=False
    )  # Make sure the antenna_id order is correct.

    # ['OFFSET', 'POSITION', 'DISH_DIAMETER', 'FLAG_ROW', 'MOUNT', 'NAME', 'STATION']
    ant_xds = ant_xds.assign_coords({"cartesian_pos_label": ["x", "y", "z"]})

    ant_xds = convert_generic_xds_to_xradio_schema(
        generic_ant_xds, ant_xds, to_new_data_variables, to_new_coords
    )

    ant_xds["ANTENNA_DISH_DIAMETER"].attrs.update({"units": ["m"], "type": "quantity"})

    ant_xds["ANTENNA_POSITION"].attrs["coordinate_system"] = "geocentric"
    ant_xds["ANTENNA_POSITION"].attrs["origin_object_name"] = "earth"

    if telescope_name in ["ALMA", "VLA", "NOEMA", "EVLA"]:
        # antenna_name = ant_xds["antenna_name"].values + "_" + ant_xds["station"].values
        # works on laptop but fails in github test runner with error:
        # numpy.core._exceptions._UFuncNoLoopError: ufunc 'add' did not contain a loop with signature matching types (dtype('<U4'), dtype('<U4')) -> None

        # Also doesn't work on github test runner:
        # antenna_name = ant_xds["antenna_name"].values
        # antenna_name = np._core.defchararray.add(antenna_name, "_")
        # antenna_name = np._core.defchararray.add(
        #     antenna_name,
        #     ant_xds["station"].values,
        # )

        # None of the native numpy functions work on the github test runner.
        antenna_name = ant_xds["antenna_name"].values
        station = ant_xds["station"].values
        antenna_name = np.array(
            list(map(lambda x, y: x + "_" + y, antenna_name, station))
        )

        ant_xds["antenna_name"] = xr.DataArray(antenna_name, dims=["antenna_name"])
        ant_xds.attrs["relocatable_antennas"] = True
    else:
        ant_xds.attrs["relocatable_antennas"] = False

    ant_xds = ant_xds.assign_coords(
        {
            "telescope_name": (
                "antenna_name",
                np.array([telescope_name for ant in ant_xds["antenna_name"]]),
            )
        }
    )

    return ant_xds


def extract_feed_info(
    ant_xds: xr.Dataset,
    in_file: str,
    antenna_id: list,
    feed_id: int,
    spectral_window_id: int,
) -> xr.Dataset:
    """
    Reformats MSv2 Feed table content to MSv4 schema.

    Parameters
    ----------
    ant_xds : xr.Dataset
        Xarray Dataset containing antenna information.
    in_file : str
        Path to the input MSv2.
    antenna_id : list
        List of antenna IDs.
    feed_id : int
        Feed ID.
    spectral_window_id : int
        Spectral window ID.

    Returns
    -------
    xr.Dataset
        Dataset updated to contain the feed information.
    """

    # Extract feed information
    generic_feed_xds = load_generic_table(
        in_file,
        "FEED",
        rename_ids=subt_rename_ids["FEED"],
        taql_where=f" where (ANTENNA_ID IN [{','.join(map(str, ant_xds.antenna_id.values))}]) AND (FEED_ID IN [{','.join(map(str, feed_id))}])",
    )  # Some Lofar and MeerKAT data have the spw column set to -1 so we can't use '(SPECTRAL_WINDOW_ID = {spectral_window_id})'

    if not generic_feed_xds:
        # Some MSv2 have a FEED table that does not cover all antenna_id (and feed_id)
        return ant_xds

    feed_spw = np.unique(generic_feed_xds.SPECTRAL_WINDOW_ID)
    if len(feed_spw) == 1 and feed_spw[0] == -1:
        generic_feed_xds = generic_feed_xds.isel(SPECTRAL_WINDOW_ID=0, drop=True)
    else:
        if spectral_window_id not in feed_spw:
            # For some spw the feed table is empty (this is the case with ALMA spw WVR#NOMINAL).
            return ant_xds
        else:
            generic_feed_xds = generic_feed_xds.sel(
                SPECTRAL_WINDOW_ID=spectral_window_id, drop=True
            )

    assert len(generic_feed_xds.TIME) == len(
        antenna_id
    ), "Can only process feed table with a single time entry for an feed, antenna and spectral_window_id."
    generic_feed_xds = generic_feed_xds.sel(
        ANTENNA_ID=antenna_id, drop=False
    )  # Make sure the antenna_id is in the same order as the xds.

    num_receptors = np.ravel(generic_feed_xds.NUM_RECEPTORS)
    num_receptors = unique_1d(num_receptors[~np.isnan(num_receptors)])

    assert (
        len(num_receptors) == 1
    ), "The number of receptors must be constant in feed table."

    to_new_data_variables = {
        "RECEPTOR_ANGLE": [
            "ANTENNA_RECEPTOR_ANGLE",
            ["antenna_name", "receptor_label"],
        ],
        "FOCUS_LENGTH": [
            "ANTENNA_FOCUS_LENGTH",
            ["antenna_name"],
        ],  # optional
    }

    to_new_coords = {
        "POLARIZATION_TYPE": ["polarization_type", ["antenna_name", "receptor_label"]]
    }

    ant_xds = convert_generic_xds_to_xradio_schema(
        generic_feed_xds,
        ant_xds,
        to_new_data_variables,
        to_new_coords=to_new_coords,
    )

    # coords["receptor_label"] = "pol_" + np.arange(ant_xds.sizes["receptor_label"]).astype(str) #Works on laptop but fails in github test runner.
    coords = {
        "receptor_label": np.array(
            list(
                map(
                    lambda x, y: x + "_" + y,
                    ["pol"] * ant_xds.sizes["receptor_label"],
                    np.arange(ant_xds.sizes["receptor_label"]).astype(str),
                )
            ),
            dtype=str,
        )
    }

    ant_xds = ant_xds.assign_coords(coords)

    # Correct to expected types. Some ALMA-SD (at least) leave receptor_label, polarization_type columns
    # in the MS empty, causing a type mismatch
    if (
        "polarization_type" in ant_xds.coords
        and ant_xds.coords["polarization_type"].dtype != str
    ):
        ant_xds.coords["polarization_type"] = ant_xds.coords[
            "polarization_type"
        ].astype(str)
    return ant_xds


def create_gain_curve_xds(
    in_file: str, spectral_window_id: int, ant_xds: xr.Dataset
) -> xr.Dataset:
    """
    Produces a gain_curve_xds, reformats MSv2 GAIN CURVE table content to MSv4 schema.

    Parameters
    ----------
    in_file : str
        Path to the input MSv2.
    spectral_window_id : int
        The ID of the spectral window.
    ant_xds : xr.Dataset
        The antenna_xds that has information such as names, stations, etc., for coordinates

    Returns
    -------
    xr.Dataset
        The updated antenna dataset with gain curve information.
    """

    gain_curve_xds = None
    if not table_exists(os.path.join(in_file, "GAIN_CURVE")):
        return gain_curve_xds

    generic_gain_curve_xds = load_generic_table(
        in_file,
        "GAIN_CURVE",
        taql_where=f" where (ANTENNA_ID IN [{','.join(map(str,ant_xds.antenna_id.values))}]) AND (SPECTRAL_WINDOW_ID = {spectral_window_id})",
    )

    if not generic_gain_curve_xds.data_vars:
        # Some times the gain_curve table is empty (this is the case with ngEHT simulation data we have).
        return gain_curve_xds

    assert (
        len(generic_gain_curve_xds.SPECTRAL_WINDOW_ID) == 1
    ), "Only one spectral window is supported."
    generic_gain_curve_xds = generic_gain_curve_xds.isel(
        SPECTRAL_WINDOW_ID=0, drop=True
    )  # Drop the spectral window dimension as it is singleton.

    assert (
        len(generic_gain_curve_xds.TIME) == 1
    ), "Only one gain curve measurement per antenna is supported."
    measured_time = generic_gain_curve_xds.coords["TIME"].values[0]
    generic_gain_curve_xds = generic_gain_curve_xds.isel(TIME=0, drop=True)

    generic_gain_curve_xds = generic_gain_curve_xds.sel(
        ANTENNA_ID=ant_xds.antenna_id, drop=False
    )  # Make sure the antenna_id is in the same order as the xds .

    gain_curve_xds = xr.Dataset(attrs={"type": "gain_curve"})

    to_new_data_variables = {
        "INTERVAL": ["GAIN_CURVE_INTERVAL", ["antenna_name"]],
        "GAIN": [
            "GAIN_CURVE",
            ["antenna_name", "poly_term", "receptor_label"],
        ],
        "SENSITIVITY": [
            "GAIN_CURVE_SENSITIVITY",
            ["antenna_name", "receptor_label"],
        ],
    }

    to_new_coords = {
        "TYPE": ["gain_curve_type", ["antenna_name"]],
    }

    gain_curve_xds = convert_generic_xds_to_xradio_schema(
        generic_gain_curve_xds,
        gain_curve_xds,
        to_new_data_variables,
        to_new_coords,
    )

    ant_borrowed_coords = {
        "antenna_name": ant_xds.coords["antenna_name"],
        "station": ant_xds.coords["station"],
        "mount": ant_xds.coords["mount"],
        "telescope_name": ant_xds.coords["telescope_name"],
        "receptor_label": ant_xds.coords["receptor_label"],
        "polarization_type": ant_xds.coords["polarization_type"],
    }
    gain_curve_xds = gain_curve_xds.assign_coords(ant_borrowed_coords)

    gain_curve_xds.attrs.update(
        {
            "measured_date": np.datetime_as_string(
                convert_casacore_time([measured_time])[0]
            )
        }
    )

    # correct expected types (for example "GAIN_CURVE" can be float32)
    for data_var in gain_curve_xds:
        if gain_curve_xds.data_vars[data_var].dtype != np.float64:
            gain_curve_xds[data_var] = gain_curve_xds[data_var].astype(np.float64)

    return gain_curve_xds


def create_phase_calibration_xds(
    in_file: str,
    spectral_window_id: int,
    ant_xds: xr.Dataset,
    time_min_max: Tuple[np.float64, np.float64],
    phase_cal_interp_time: Union[xr.DataArray, None] = None,
) -> xr.Dataset:
    """
    Produces a phase_calibration_xds, reformats MSv2 Phase Cal table content to MSv4 schema.

    Parameters
    ----------
    in_file : str
        Path to the input MSv2.
    spectral_window_id : int
        The ID of the spectral window.
    ant_xds : xr.Dataset
        The antenna_xds that has information such as names, stations, etc., for coordinates
    time_min_max : Tuple[np.float46, np.float64]
        Min / max times to constrain loading (usually to the time range relevant to an MSv4)
    interp_time : Union[xr.DataArray, None]
        Time axis to interpolate the data vars to (usually main MSv4 time)

    Returns
    -------
    xr.Dataset
        The updated antenna dataset with phase cal information.
    """

    phase_cal_xds = None
    if not table_exists(os.path.join(in_file, "PHASE_CAL")):
        return phase_cal_xds

    # Only read data between the min and max times of the visibility data in the MSv4.
    taql_time_range = make_taql_where_between_min_max(
        time_min_max, in_file, "PHASE_CAL", "TIME"
    )
    generic_phase_cal_xds = load_generic_table(
        in_file,
        "PHASE_CAL",
        timecols=["TIME"],
        taql_where=f" {taql_time_range} AND (ANTENNA_ID IN [{','.join(map(str,ant_xds.antenna_id.values))}]) AND (SPECTRAL_WINDOW_ID = {spectral_window_id})",
    )

    assert (
        len(generic_phase_cal_xds.SPECTRAL_WINDOW_ID) == 1
    ), "Only one spectral window is supported."
    generic_phase_cal_xds = generic_phase_cal_xds.isel(
        SPECTRAL_WINDOW_ID=0, drop=True
    )  # Drop the spectral window dimension as it is singleton.

    generic_phase_cal_xds = generic_phase_cal_xds.sel(
        ANTENNA_ID=ant_xds.antenna_id, drop=False
    )  # Make sure the antenna_id is in the same order as the xds.

    to_new_data_variables = {
        "INTERVAL": ["PHASE_CAL_INTERVAL", ["antenna_name", "time_phase_cal"]],
        "TONE_FREQUENCY": [
            "PHASE_CAL_TONE_FREQUENCY",
            ["antenna_name", "time_phase_cal", "tone_label", "receptor_label"],
        ],
        "PHASE_CAL": [
            "PHASE_CAL",
            ["antenna_name", "time_phase_cal", "tone_label", "receptor_label"],
        ],
        "CABLE_CAL": ["PHASE_CAL_CABLE_CAL", ["antenna_name", "time_phase_cal"]],
    }

    to_new_coords = {
        "TIME": ["time_phase_cal", ["time_phase_cal"]],
    }

    phase_cal_xds = xr.Dataset(attrs={"type": "phase_calibration"})
    phase_cal_xds = convert_generic_xds_to_xradio_schema(
        generic_phase_cal_xds, phase_cal_xds, to_new_data_variables, to_new_coords
    )

    phase_cal_xds["PHASE_CAL"] = phase_cal_xds["PHASE_CAL"].transpose(
        "antenna_name", "time_phase_cal", "receptor_label", "tone_label"
    )
    phase_cal_xds["PHASE_CAL_TONE_FREQUENCY"] = phase_cal_xds[
        "PHASE_CAL_TONE_FREQUENCY"
    ].transpose("antenna_name", "time_phase_cal", "receptor_label", "tone_label")

    ant_borrowed_coords = {
        "antenna_name": ant_xds.coords["antenna_name"],
        "station": ant_xds.coords["station"],
        "mount": ant_xds.coords["mount"],
        "telescope_name": ant_xds.coords["telescope_name"],
        "receptor_label": ant_xds.coords["receptor_label"],
        "polarization_type": ant_xds.coords["polarization_type"],
    }
    # phase_cal_xds = phase_cal_xds.assign_coords({"tone_label" : "freq_" + np.arange(phase_cal_xds.sizes["tone_label"]).astype(str)}) #Works on laptop but fails in github test runner.
    tone_label_coord = {
        "tone_label": np.array(
            list(
                map(
                    lambda x, y: x + "_" + y,
                    ["freq"] * phase_cal_xds.sizes["tone_label"],
                    np.arange(phase_cal_xds.sizes["tone_label"]).astype(str),
                )
            )
        )
    }
    phase_cal_xds = phase_cal_xds.assign_coords(ant_borrowed_coords | tone_label_coord)

    # Adjust expected types
    phase_cal_xds["time_phase_cal"] = (
        phase_cal_xds.time_phase_cal.astype("float64").astype("float64") / 10**9
    )

    phase_cal_xds = rename_and_interpolate_to_time(
        phase_cal_xds, "time_phase_cal", phase_cal_interp_time, "phase_cal_xds"
    )

    return phase_cal_xds



from astropy import units as u
import copy
import os
from typing import List, Tuple, Union

import dask
import dask.array as da
import toolviper.utils.logger as logger
import numpy as np
import xarray as xr
from astropy import units as u
from casacore import tables
from casacore.images import coordinates

from .common import (
    _active_mask,
    _object_name,
    _open_image_ro,
    _pointing_center,
)
from ..common import (
    _compute_linear_world_values,
    _compute_velocity_values,
    _compute_world_sph_dims,
    _convert_beam_to_rad,
    _default_freq_info,
    _doppler_types,
    _get_unit,
    _image_type,
    _l_m_attr_notes,
)
from ...._utils._casacore.tables import extract_table_attributes, open_table_ro
from xradio._utils.coord_math import _deg_to_rad
from xradio._utils.dict_helpers import make_quantity

"""
def _add_coord_attrs(xds: xr.Dataset, icoords: dict, dir_axes: list) -> xr.Dataset:
    _add_time_attrs(xds, icoords)
    _add_freq_attrs(xds, icoords)
    xds = _add_vel_attrs(xds, icoords)
    xds = _add_lin_attrs(xds, icoords, dir_axes)
    return xds
"""


def _add_lin_attrs(xds, coord_dict, dir_axes):
    for k in coord_dict:
        if k.startswith("linear"):
            ld = coord_dict[k]
            for i in (0, 1):
                meta = {}
                meta["units"] = ld["units"][i]
                meta["crval"] = ld["crval"][i]
                meta["cdelt"] = ld["cdelt"][i]
                xds[dir_axes[i]].attrs = copy.deepcopy(meta)
            break
    return xds


def _add_freq_attrs(xds, coord_dict):
    freq_coord = xds["frequency"]
    meta = {}
    for k in coord_dict:
        if k.startswith("spectral"):
            sd = coord_dict[k]
            meta["rest_frequency"] = make_quantity(sd["restfreq"], "Hz")
            meta["type"] = "frequency"
            meta["units"] = sd["unit"]
            meta["frame"] = sd["system"]
            meta["wave_unit"] = sd["waveUnit"]
            meta["crval"] = sd["wcs"]["crval"]
            meta["cdelt"] = sd["wcs"]["cdelt"]
    if not meta:
        # this is the default frequency information CASA creates
        meta = _default_freq_info()
    freq_coord.attrs = meta


def _add_mask(
    xds: xr.Dataset, name: str, ary: Union[np.ndarray, da.array], dimorder: list
) -> xr.Dataset:
    xda = xr.DataArray(ary, dims=dimorder)
    # True pixels are good in numpy masked arrays
    xda = da.logical_not(xda)
    xda.attrs["image_type"] = "Mask"
    xda = xda.rename(name)
    xds[xda.name] = xda
    return xds


def _add_sky_or_aperture(
    xds: xr.Dataset,
    ary: Union[np.ndarray, da.array],
    dimorder: list,
    img_full_path: str,
    has_sph_dims: bool,
) -> xr.Dataset:
    xda = xr.DataArray(ary, dims=dimorder).astype(ary.dtype)
    with _open_image_ro(img_full_path) as casa_image:
        image_type = casa_image.info()["imageinfo"]["imagetype"]
        unit = casa_image.unit()
    xda.attrs[_image_type] = image_type
    xda.attrs["units"] = unit
    name = "SKY" if has_sph_dims else "APERTURE"
    xda = xda.rename(name)
    xds[xda.name] = xda
    return xds


def _get_time_format(value: float, unit: str) -> str:
    if value >= 40000 and value <= 100000 and unit == "d":
        return "MJD"
    else:
        return ""


def _add_time_attrs(xds: xr.Dataset, coord_dict: dict) -> xr.Dataset:
    # time_coord = xds['time']
    meta = {}
    meta["type"] = "time"
    meta["scale"] = coord_dict["obsdate"]["refer"]
    meta["units"] = coord_dict["obsdate"]["m0"]["unit"]
    meta["format"] = _get_time_format(xds["time"][0], meta["units"])
    xds["time"].attrs = copy.deepcopy(meta)
    # xds['time'] = time_coord
    return xds


def _add_vel_attrs(xds: xr.Dataset, coord_dict: dict) -> xr.Dataset:
    vel_coord = xds["velocity"]
    meta = {"units": "m/s"}
    for k in coord_dict:
        if k.startswith("spectral"):
            sd = coord_dict[k]
            meta["doppler_type"] = _doppler_types[sd["velType"]]
            break
    if not meta:
        meta["doppler_type"] = _doppler_types[0]
    meta["type"] = "doppler"
    vel_coord.attrs = copy.deepcopy(meta)
    xds["velocity"] = vel_coord
    return xds


def _casa_image_to_xds_attrs(img_full_path: str, history: bool = True) -> dict:
    """
    Get the xds level attribut/es as a python dictionary
    """
    with _open_image_ro(img_full_path) as casa_image:
        meta_dict = casa_image.info()
    coord_dict = copy.deepcopy(meta_dict["coordinates"])
    attrs = {}
    dir_key = None
    for k in coord_dict.keys():
        if k.startswith("direction"):
            dir_key = k
            break
    if dir_key:
        # shared direction coordinate attributes
        coord_dir_dict = copy.deepcopy(coord_dict[dir_key])
        system = "system"
        if system not in coord_dir_dict:
            raise RuntimeError("No direction reference frame found")
        casa_system = coord_dir_dict[system]
        ap_system, ap_equinox = _convert_direction_system(casa_system, "native")
        dir_dict = {}
        dir_dict["reference"] = {
            "frame": ap_system,
            "type": "sky_coord",
            "equinox": ap_equinox if ap_equinox else None,
            "value": [0.0, 0.0],
            "units": ["rad", "rad"],
        }
        for i in range(2):
            unit = u.Unit(_get_unit(coord_dir_dict["units"][i]))
            q = coord_dir_dict["crval"][i] * unit
            x = q.to("rad")
            dir_dict["reference"]["value"][i] = x.value
        k = "latpole"
        if k in coord_dir_dict:
            for j in (k, "longpole"):
                dir_dict[j] = make_quantity(coord_dir_dict[j] * _deg_to_rad, "rad")
        for j in ("pc", "projection_parameters", "projection"):
            if j in coord_dir_dict:
                dir_dict[j] = coord_dir_dict[j]
        attrs["direction"] = dir_dict
    attrs["telescope"] = {}
    telescope = attrs["telescope"]
    attrs["obsdate"] = {"type": "time"}
    obsdate = attrs["obsdate"]
    attrs[_pointing_center] = coord_dict["pointingcenter"].copy()
    for k in ("observer", "obsdate", "telescope", "telescopeposition"):
        if k.startswith("telescope"):
            if k == "telescope":
                telescope["name"] = coord_dict[k]
            elif k in coord_dict:
                telescope["position"] = coord_dict[k]
                telescope["position"]["ellipsoid"] = telescope["position"]["refer"]
                if telescope["position"]["refer"] == "ITRF":
                    telescope["position"]["ellipsoid"] = "GRS80"
                telescope["position"]["units"] = [
                    telescope["position"]["m0"]["unit"],
                    telescope["position"]["m1"]["unit"],
                    telescope["position"]["m2"]["unit"],
                ]
                telescope["position"]["value"] = [
                    telescope["position"]["m0"]["value"],
                    telescope["position"]["m1"]["value"],
                    telescope["position"]["m2"]["value"],
                ]
                del (
                    telescope["position"]["refer"],
                    telescope["position"]["m0"],
                    telescope["position"]["m1"],
                    telescope["position"]["m2"],
                )
        elif k == "obsdate":
            obsdate["scale"] = coord_dict[k]["refer"]
            obsdate["units"] = coord_dict[k]["m0"]["unit"]
            obsdate["value"] = coord_dict[k]["m0"]["value"]
            obsdate["format"] = _get_time_format(obsdate["value"], obsdate["units"])
            obsdate["type"] = "time"
        else:
            attrs[k] = coord_dict[k] if k in coord_dict else ""
    imageinfo = meta_dict["imageinfo"]
    obj = "objectname"
    attrs[_object_name] = imageinfo[obj] if obj in imageinfo else ""
    attrs["beam"] = _get_beam(imageinfo)
    attrs["user"] = meta_dict["miscinfo"]
    defmask = "Image_defaultmask"
    with open_table_ro(img_full_path) as casa_table:
        attrs[_active_mask] = (
            casa_table.getkeyword(defmask)
            if defmask in casa_table.keywordnames()
            else None
        )
    attrs["description"] = None
    # if also loading history, put it as another xds in the attrs
    if history:
        htable = os.sep.join([img_full_path, "logtable"])
        if os.path.isdir(htable):
            attrs["history"] = read_generic_table(htable)
        else:
            logger.warning(
                f"Unable to find history table {htable}. History will not be included"
            )
    return copy.deepcopy(attrs)


def _casa_image_to_xds_coords(
    img_full_path: str, verbose: bool, do_sky_coords: bool
) -> dict:
    """
    TODO: complete documentation
    Create an xds without any pixel data from metadata from the specified CASA image
    """
    attrs = {}
    # casa_image = images.image(img_full_path)
    with _open_image_ro(img_full_path) as casa_image:
        # shape list is the reverse of the actual image shape
        shape = casa_image.shape()[::-1]
        meta_dict = casa_image.info()
        csys = casa_image.coordinates()
    axis_names = _flatten_list(csys.get_axes())[::-1]
    coord_dict = meta_dict["coordinates"]
    attrs["icoords"] = coord_dict
    diraxes = [
        aa.lower().replace(" ", "_")
        for cc in coord_dict.items()
        if (cc[0][:-1] in ["direction", "linear"]) and len(cc[1]["axes"]) >= 2
        for aa in cc[1]["axes"]
    ]
    attrs["dir_axes"] = diraxes
    dimmap = _get_dimmap(coord_dict, verbose)
    attrs["dimmap"] = dimmap
    sphr_dims = (
        [dimmap["l"], dimmap["m"]] if ("l" in dimmap) and ("m" in dimmap) else []
    )
    attrs["sphr_dims"] = sphr_dims
    coords = {}
    coord_attrs = {}
    (coords["time"], coord_attrs["time"]) = _get_time_values_attrs(coord_dict)
    (coords["polarization"], coord_attrs["polarization"]) = _get_pol_values_attrs(
        coord_dict
    )
    (coords["frequency"], coord_attrs["frequency"]) = _get_freq_values_attrs(
        csys, shape
    )
    (velocity_vals, coord_attrs["velocity"]) = _get_velocity_values_attrs(
        coord_dict, coords["frequency"]
    )
    coords["velocity"] = (["frequency"], velocity_vals)
    if len(sphr_dims) > 0:
        crpix = _flatten_list(csys.get_referencepixel())[::-1]
        inc = _flatten_list(csys.get_increment())[::-1]
        unit = _flatten_list(csys.get_unit())[::-1]
        attr_note = _l_m_attr_notes()
        for c in ["l", "m"]:
            idx = dimmap[c]
            delta = ((inc[idx]) * u.Unit(_get_unit(unit[idx]))).to("rad").value
            coords[c] = _compute_linear_world_values(
                naxis=shape[idx], crval=0.0, crpix=crpix[idx], cdelt=delta
            )
            coord_attrs[c] = {
                "crval": 0.0,
                "cdelt": delta,
                "units": "rad",
                "type": "quantity",
                "note": attr_note[c],
            }
        if do_sky_coords:
            for k in coord_dict.keys():
                if k.startswith("direction"):
                    dc = coordinates.directioncoordinate(coord_dict[k])
                    break
            crval = _flatten_list(csys.get_referencevalue())[::-1]
            pick = lambda my_list: [my_list[i] for i in sphr_dims]
            my_ret = _compute_world_sph_dims(
                projection=dc.get_projection(),
                shape=pick(shape),
                ctype=diraxes,
                crval=pick(crval),
                crpix=pick(crpix),
                cdelt=pick(inc),
                cunit=pick(unit),
            )
            for i in [0, 1]:
                axis_name = my_ret["axis_name"][i]
                coords[axis_name] = (["l", "m"], my_ret["value"][i])
                coord_attrs[axis_name] = {}
    else:
        # Fourier image
        ret = _get_uv_values_attrs(coord_dict, axis_names, shape)
        for z in ["u", "v"]:
            coords[z], coord_attrs[z] = ret[z]
    attrs["shape"] = shape
    xds = xr.Dataset(coords=coords)
    for c in coord_attrs.keys():
        xds[c].attrs = coord_attrs[c]
    attrs["xds"] = xds
    return attrs


def _convert_direction_system(
    casa_system: str, which: str, verbose: bool = True
) -> tuple:
    if casa_system == "J2000":
        if verbose:
            logger.info(
                f"J2000 found as {which} reference frame in CASA image "
                'This corresponds to FK5(equinox="J2000") in astropy. '
                "Metadata will be written appropriately"
            )
        return ("FK5", "J2000.0")
    elif casa_system == "B1950":
        if verbose:
            logger.info(
                f"B1950 found as {which} reference frame in CASA image "
                'This corresponds to FK4(equinox="B1950") in astropy. '
                "Metadata will be written appropriately"
            )
        return ("FK4", "B1950.0")
    elif casa_system in ("GALACTIC", "ICRS"):
        return (casa_system.lower(), None)
    else:
        raise Exception(
            f"astropy does not support frame {casa_system} and this "
            "application does not support converting it to "
            "something astropy does support. You can try to regridding "
            "it to ICRS, GALACTIC, J2000, or B1950 in CASA and then "
            "re-run this application on the regridded image"
        )


def _flatten_list(list_of_lists: list) -> list:
    flat = []
    for x in list_of_lists:
        if type(x) == list or type(x) == np.ndarray:
            flat.extend(_flatten_list(x))
        else:
            flat.append(x)
    return flat


def _get_beam(imageinfo: dict):
    """Returns None if no beam. Multiple beams are handled elsewhere"""
    k = "restoringbeam"
    if k in imageinfo and "major" in imageinfo[k]:
        return _convert_beam_to_rad(imageinfo[k])
    return None


def _get_chunk_list(
    chunks: dict, coords: list, image_shape: Union[list, tuple]
) -> tuple:
    ret_list = list(image_shape)
    axis = 0
    for c in coords:
        if c == "direction" or c == "linear":
            lm = ("l", "m")
            uv = ("u", "v")
            for i in (0, 1):
                for k in (lm[i], uv[i]):
                    if k in chunks:
                        ret_list[axis] = chunks[k]
                        break
                # add an axis because direction has 2 axes
                if i == 0:
                    axis += 1
        elif c == "spectral":
            if "frequency" in chunks:
                ret_list[axis] = chunks["frequency"]
        elif c == "stokes":
            if "polarization" in chunks:
                ret_list[axis] = chunks["polarization"]
        else:
            raise Exception(f"Unhandled coordinate type {c}")
        axis += 1
    return tuple(ret_list)


def _get_dimmap(coords: list, verbose: bool = False) -> dict:
    # example of dimmap:
    # [('direction0', 0), ('direction1', 1), ('spectral0', 2), ('stokes0', 3)]
    dimmap: list[tuple] = [
        (coord[:-1] + str(ii), ci)
        for coord in coords
        if coord[:-1] in ["direction", "stokes", "spectral", "linear"]
        for ii, ci in enumerate(coords["pixelmap%s" % coord[-1]])
    ]
    if verbose:
        print(f"dimmap: {dimmap}")
    # example of dimmap after next statment
    # [('l', 0), ('m', 1), ('chan', 2), ('polarization', 3)]
    dimmap = [
        (
            rr[0]
            .replace("stokes0", "polarization")
            .replace("spectral0", "chan")
            .replace("direction0", "l")
            .replace("direction1", "m")
            .replace("linear0", "u")
            .replace("linear1", "v"),
            rr[1],
        )
        for rr in dimmap
    ]
    if verbose:
        print(f"dimmap: {dimmap}")
    if ("linear0" in np.vstack(dimmap)[:, 0]) and (
        "linear1" in np.vstack(dimmap)[:, 0]
    ):
        dimmap = [
            (rr[0].replace("linear0", "l").replace("linear1", "m"), rr[1])
            for rr in dimmap
        ]
    dimmap = [
        (rr[0].replace("linear0", "component"), rr[1]) for rr in dimmap if rr[1] >= 0
    ]
    # conversion to dict, example dimmap after this statement
    # dimmap: {'l': 0, 'm': 1, 'chan': 2, 'polarization': 3}
    dimmap = dict(
        [
            (
                (diraxes[int(rr[0][-1])], rr[1])
                if rr[0].startswith("linear") or rr[0].startswith("direction")
                else rr
            )
            for rr in dimmap
        ]
    )
    if verbose:
        print(f"dimmap: {dimmap}")
    return dimmap


def _get_freq_values(coords: coordinates.coordinatesystem, shape: tuple) -> list:
    idx = _get_image_axis_order(coords)[::-1].index("Frequency")
    if idx >= 0:
        coord_dict = coords.dict()
        for k in coord_dict:
            if k.startswith("spectral"):
                wcs = coord_dict[k]["wcs"]
                return _compute_linear_world_values(
                    naxis=shape[idx],
                    crval=wcs["crval"],
                    crpix=wcs["crpix"],
                    cdelt=wcs["cdelt"],
                )
    else:
        return [1420e6]


def _get_freq_values_attrs(
    casa_coords: coordinates.coordinatesystem, shape: tuple
) -> Tuple[List[float], dict]:
    values = None
    attrs = {}
    idx = _get_image_axis_order(casa_coords)[::-1].index("Frequency")
    if idx >= 0:
        casa_coord_dict = casa_coords.dict()
        for k in casa_coord_dict:
            if k.startswith("spectral"):
                sd = casa_coord_dict[k]
                wcs = sd["wcs"]
                values = _compute_linear_world_values(
                    naxis=shape[idx],
                    crval=wcs["crval"],
                    crpix=wcs["crpix"],
                    cdelt=wcs["cdelt"],
                )
                attrs["rest_frequency"] = make_quantity(sd["restfreq"], "Hz")
                attrs["type"] = "frequency"
                attrs["units"] = sd["unit"]
                attrs["frame"] = sd["system"]
                attrs["wave_unit"] = sd["waveUnit"]
                attrs["crval"] = sd["wcs"]["crval"]
                attrs["cdelt"] = sd["wcs"]["cdelt"]
                attrs = copy.deepcopy(attrs)
                break
    else:
        values = [1420e6]
        # this is the default frequency information CASA creates
        attrs = _default_freq_info()
    return (values, attrs)


def _get_image_axis_order(coords: coordinates.coordinatesystem) -> list:
    """
    get the *reverse* order of image axes
    """
    axis_names = coords.get_axes()[::-1]
    ncoords = len(axis_names)
    csys = coords.dict()
    ordered = len(_flatten_list(axis_names)) * [""]
    for i in range(ncoords - 1, -1, -1):
        axes = csys["pixelmap" + str(i)]
        if len(axes) == 1:
            ordered[axes[0]] = axis_names[i]
        elif len(axes) == 2:
            ordered[axes[0]] = axis_names[i][1]
            ordered[axes[1]] = axis_names[i][0]
    ordered = _flatten_list(ordered)[::-1]
    return ordered


def _get_image_dim_order(coords: coordinates.coordinatesystem) -> list:
    """
    Get the xds dim order of the input image. The returned list is in casacore
    order, which is the reverse of the actual image dimension order
    """
    flat = _get_image_axis_order(coords)
    ret = []
    for axis in flat:
        b = axis.lower()
        if b.startswith("right") or b.startswith("uu"):
            ret.append("l")
        elif b.startswith("dec") or b.startswith("vv"):
            ret.append("m")
        elif b.startswith("frequency"):
            ret.append("frequency")
        elif b.startswith("stok"):
            ret.append("polarization")
        else:
            raise Exception(f"Unhandled axis name {c}")
    return ret


def _get_mask_names(infile: str) -> list:
    t = tables.table(infile)
    tb_tool = tables.table(
        infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
    )
    mymasks = t.getkeyword("masks") if "masks" in t.keywordnames() else []
    t.close()
    return mymasks


def _get_multibeam(imageinfo: dict) -> Union[np.ndarray, None]:
    """Returns None if the image does not have multiple (per-plane) beams"""
    p = "perplanebeams"
    if p not in imageinfo:
        return None
    beam = imageinfo[p]
    nchan = beam["nChannels"]
    npol = beam["nStokes"]
    beam_array = np.zeros([1, npol, nchan, 3])
    for c in range(nchan):
        for p in range(npol):
            k = nchan * p + c
            b = beam["*" + str(k)]
            beam_dict = _convert_beam_to_rad(b)
            beam_array[0][p][c][0] = beam_dict["major"]["value"]
            beam_array[0][p][c][1] = beam_dict["minor"]["value"]
            beam_array[0][p][c][2] = beam_dict["pa"]["value"]
    return beam_array


def _get_persistent_block(
    infile: str,
    shapes: tuple,
    starts: tuple,
    dimorder: list,
    transpose_list: list,
    new_axes: list,
) -> xr.DataArray:
    block = _read_image_chunk(infile, shapes, starts)
    block = np.expand_dims(block, new_axes)
    block = block.transpose(transpose_list)
    block = xr.DataArray(block, dims=dimorder)
    return block


def _get_pol_values_attrs(casa_coord_dict: dict) -> Tuple[List[str], dict]:
    values = None
    for k in casa_coord_dict:
        if k.startswith("stokes"):
            values = casa_coord_dict[k]["stokes"]
    if values is None:
        values = ["I"]
    return (values, {})


def _get_pol_values(coord_dict: dict) -> List[str]:
    for k in coord_dict:
        if k.startswith("stokes"):
            return coord_dict[k]["stokes"]
    return ["I"]


def _get_starts_shapes_slices(
    blockdes: dict, coords: coordinates.coordinatesystem, cshape: list
) -> tuple:
    img_dim_order = _get_image_dim_order(coords)
    starts = []
    shapes = []
    slices = {}
    for i, dim in enumerate(img_dim_order):
        if dim not in ["polarization", "frequency", "l", "m", "u", "v"]:
            raise RuntimeError(f"Unsupported dimension {dim}")
        if dim in blockdes:
            extent = blockdes[dim]
            if isinstance(extent, int):
                starts.append(extent)
                shapes.append(extent + 1)
                slices[dim] = slice(extent, extent + 1)
            elif isinstance(extent, slice):
                starts.append(extent.start)
                shapes.append(extent.stop - extent.start)
                slices[dim] = slice(extent.start, extent.stop)
            else:
                raise Exception(f"Unhandled extent type {type(extent)}")
        else:
            starts.append(0)
            shapes.append(cshape[i])
    return starts, shapes, slices


def _get_time_values_attrs(cimage_coord_dict: dict) -> Tuple[List[float], dict]:
    attrs = {}
    attrs["type"] = "time"
    attrs["scale"] = cimage_coord_dict["obsdate"]["refer"]
    unit = cimage_coord_dict["obsdate"]["m0"]["unit"]
    attrs["units"] = unit
    time_val = cimage_coord_dict["obsdate"]["m0"]["value"]
    attrs["format"] = _get_time_format(time_val, unit)
    return ([time_val], copy.deepcopy(attrs))


def _get_time_values(coord_dict):
    return [coord_dict["obsdate"]["m0"]["value"]]


def _get_transpose_list(coords: coordinates.coordinatesystem) -> list:
    flat = _get_image_axis_order(coords)
    transpose_list = 5 * [-1]
    # time axis
    transpose_list[0] = 4
    new_axes = [4]
    last_axis = 3
    not_covered = ["l", "m", "u", "v", "s", "f"]
    csys = coords.dict()
    for i, c in enumerate(flat):
        b = c.lower()
        if b.startswith("right") or b.startswith("uu"):
            transpose_list[3] = i
            # transpose_list[3] = csys['pixelmap0'][0]
            not_covered.remove("l")
            not_covered.remove("u")
        elif b.startswith("dec") or b.startswith("vv"):
            transpose_list[4] = i
            # transpose_list[4] = csys['pixelmap0'][1]
            not_covered.remove("m")
            not_covered.remove("v")
        elif b.startswith("frequency"):
            # transpose_list[2] = csys['pixelmap1'][0]
            transpose_list[2] = i
            not_covered.remove("f")
        elif b.startswith("stok"):
            transpose_list[1] = i
            # transpose_list[1] = csys['pixelmap2'][0]
            not_covered.remove("s")
        else:
            raise Exception(f"Unhandled axis name {c}")
    h = {"l": 3, "m": 4, "u": 3, "v": 4, "f": 2, "s": 1}
    for p in not_covered:
        transpose_list[h[p]] = last_axis
        new_axes.append(last_axis)
        last_axis -= 1
    new_axes.sort()
    if transpose_list.count(-1) > 0:
        raise Exception(f"Logic error: axes {axes}, transpose_list {transpose_list}")
    return transpose_list, new_axes


def _get_uv_values(
    coord_dict: dict, axis_names: List[str], shape: Tuple[int]
) -> Tuple[List[float]]:
    for i, axis in enumerate(["UU", "VV"]):
        idx = axis_names.index(axis)
        if idx >= 0:
            for k in coord_dict:
                cdict = coord_dict[k]
                if k.startswith("linear"):
                    z = []
                    crpix = cdict["crpix"][i]
                    crval = cdict["crval"][i]
                    cdelt = cdict["cdelt"][i]
                    for i in range(shape[idx]):
                        f = (i - crpix) * cdelt + crval
                        z.append(f)
                    if axis == "UU":
                        u = z
                    else:
                        v = z
    return u, v


def _get_uv_values_attrs(
    casa_coord_dict: dict, axis_names: List[str], shape: Tuple[int]
) -> dict:
    ret = {}
    for i, axis in enumerate(["UU", "VV"]):
        idx = axis_names.index(axis)
        if idx >= 0:
            for k in casa_coord_dict:
                cdict = casa_coord_dict[k]
                if k.startswith("linear"):
                    crpix = cdict["crpix"][i]
                    crval = cdict["crval"][i]
                    cdelt = cdict["cdelt"][i]
                    attrs = {
                        "crval": crval,
                        "cdelt": cdelt,
                        "units": cdict["units"][i],
                        "type": "quantity",
                    }
                    z = _compute_linear_world_values(
                        naxis=shape[idx],
                        crpix=crpix,
                        crval=crval,
                        cdelt=cdelt,
                    )
                    name = "u" if axis == "UU" else "v"
                    ret[name] = (z, attrs)
    return ret


def _get_velocity_values(coord_dict: dict, freq_values: list) -> list:
    restfreq = 1420405751.786
    for k in coord_dict:
        if k.startswith("spectral"):
            restfreq = coord_dict[k]["restfreq"]
            break
    return _compute_velocity_values(
        restfreq=restfreq, freq_values=freq_values, doppler="RADIO"
    )


def _get_velocity_values_attrs(
    coord_dict: dict, freq_values: List[float]
) -> Tuple[List[float], dict]:
    restfreq = 1420405751.786
    attrs = {}
    for k in coord_dict:
        if k.startswith("spectral"):
            sd = coord_dict[k]
            restfreq = sd["restfreq"]
            attrs["doppler_type"] = _doppler_types[sd["velType"]]
            break
    if not attrs:
        attrs["doppler_type"] = _doppler_types[0]

    attrs["units"] = "m/s"
    attrs["type"] = "doppler"
    return (
        _compute_velocity_values(
            restfreq=restfreq, freq_values=freq_values, doppler="RADIO"
        ),
        copy.deepcopy(attrs),
    )


def _multibeam_array(
    xds: xr.Dataset, img_full_path: str, as_dask_array: bool
) -> Union[xr.DataArray, None]:
    """This should only be called after the xds.beam attr has been set"""
    if xds.attrs["beam"] is None:
        # the image may have multiple beams
        with _open_image_ro(img_full_path) as casa_image:
            imageinfo = casa_image.info()["imageinfo"]
        mb = _get_multibeam(imageinfo)
        if mb is not None:
            # multiple beams are stored as a data varialbe, so remove
            # the beam xds attr
            del xds.attrs["beam"]
            if as_dask_array:
                mb = da.array(mb)
            xdb = xr.DataArray(
                mb, dims=["time", "polarization", "frequency", "beam_param"]
            )
            xdb = xdb.rename("beam")
            xdb = xdb.assign_coords(beam_param=["major", "minor", "pa"])
            xdb.attrs["units"] = "rad"
            return xdb
    else:
        return None


###########################################################################
##
## read_generic_table() - read casacore table into memory resident xds
##
###########################################################################
def read_generic_table(infile, subtables=False, timecols=None, ignore=None):
    """
    read generic casacore table format to xarray dataset loaded in memory

    Parameters
    ----------
    infile : str
        Input table filename. To read a subtable simply append the subtable folder name under the main table (ie infile = '/path/mytable.tbl/mysubtable')
    subtables : bool
        Whether or not to include subtables underneath the specified table. If true, an attribute called subtables will be added to the returned xds.
        Default False
    timecols : list
        list of column names to convert to numpy datetime format. Default None leaves times as their original casacore format.
    ignore : list
        list of column names to ignore and not try to read. Default None reads all columns

    Returns
    -------
    xarray.core.dataset.Dataset
    """
    if timecols is None:
        timecols = []
    if ignore is None:
        ignore = []

    infile = os.path.expanduser(infile)
    assert os.path.isdir(infile), "invalid input filename to read_generic_table"

    attrs = extract_table_attributes(infile)
    with open_table_ro(infile) as tb_tool:
        if tb_tool.nrows() == 0:
            return xr.Dataset(attrs=attrs)

        dims = ["row"] + ["d%i" % ii for ii in range(1, 20)]
        cols = tb_tool.colnames()
        ctype = dict(
            [
                (col, tb_tool.getcell(col, 0))
                for col in cols
                if ((col not in ignore) and (tb_tool.iscelldefined(col, 0)))
            ]
        )
        mvars, mcoords, xds = {}, {}, xr.Dataset()

        tr = tb_tool.row(ignore, exclude=True)[:]

        # extract data for each col
        for col in ctype.keys():
            if tb_tool.coldatatype(col) == "record":
                continue  # not supported

            try:
                data = np.stack([rr[col] for rr in tr])  # .astype(ctype[col].dtype)
                if isinstance(tr[0][col], dict):
                    data = np.stack(
                        [
                            (
                                rr[col]["array"].reshape(rr[col]["shape"])
                                if len(rr[col]["array"]) > 0
                                else np.array([""])
                            )
                            for rr in tr
                        ]
                    )
            except:
                # sometimes the columns are variable, so we need to standardize to the largest sizes
                if len(np.unique([isinstance(rr[col], dict) for rr in tr])) > 1:
                    continue  # can't deal with this case
                mshape = np.array(max([np.array(rr[col]).shape for rr in tr]))
                try:
                    data = np.stack(
                        [
                            np.pad(
                                (
                                    rr[col]
                                    if len(rr[col]) > 0
                                    else np.array(rr[col]).reshape(
                                        np.arange(len(mshape)) * 0
                                    )
                                ),
                                [(0, ss) for ss in mshape - np.array(rr[col]).shape],
                                "constant",
                                constant_values=np.array([np.nan]).astype(
                                    np.array(ctype[col]).dtype
                                )[0],
                            )
                            for rr in tr
                        ]
                    )
                except:
                    data = []

            if len(data) == 0:
                continue
            if col in timecols:
                convert_time(data)
            if col.endswith("_ID"):
                mcoords[col] = xr.DataArray(
                    data,
                    dims=[
                        "d%i_%i" % (di, ds)
                        for di, ds in enumerate(np.array(data).shape)
                    ],
                )
            else:
                mvars[col] = xr.DataArray(
                    data,
                    dims=[
                        "d%i_%i" % (di, ds)
                        for di, ds in enumerate(np.array(data).shape)
                    ],
                )

        xds = xr.Dataset(mvars, coords=mcoords)
        xds = xds.rename(dict([(dv, dims[di]) for di, dv in enumerate(xds.sizes)]))
        attrs["bad_cols"] = list(
            np.setdiff1d(
                [dv for dv in tb_tool.colnames()],
                [dv for dv in list(xds.data_vars) + list(xds.coords)],
            )
        )

        # if this table has subtables, use a recursive call to store them in subtables attribute
        if subtables:
            stbl_list = sorted(
                [
                    tt
                    for tt in os.listdir(infile)
                    if os.path.isdir(os.path.join(infile, tt))
                    and tables.tableexists(os.path.join(infile, tt))
                ]
            )
            attrs["subtables"] = []
            for ii, subtable in enumerate(stbl_list):
                sxds = read_generic_table(
                    os.path.join(infile, subtable),
                    subtables=subtables,
                    timecols=timecols,
                    ignore=ignore,
                )
                if len(sxds.sizes) != 0:
                    attrs["subtables"] += [(subtable, sxds)]
        xds = xds.assign_attrs(attrs)
    return xds


def _read_image_array(
    infile: str,
    chunks: Union[list, dict],
    mask: str = None,
    verbose: bool = False,
    blc=None,
    trc=None,
) -> dask.array:
    """
    Read an array of image pixels into a dask array. The returned dask array
    will have axes in time, polarization, frequency, l, m order
    If specified, it's the caller's responsibility to ensure the specified blc and
    trc are coincident with chunk corners. If not, there could be performance degradation
    The blc, trc box is inclusive of blc pixel coordinates and exclusive trc pixel
    coordinates, which mimics the behavior of numpy array slicing
    ie blc = [0, 0, 0] and trc = [1023, 1023, 1023] excludes pixel [1023, 1023, 1023]
    and has shape [1023, 1023, 1023]
    blc and trc can be any array type object
    blc and trc are taken wrt the CASA image. Not the casacore image in which they
    would be reversed
    :param infile: Path to the input CASA image
    :type infile: str, required
    :param chunks: The desired dask chunk size.
                   If list, the ordering is based on the ordering of the input
                   axes. So, for example, if the input image has axes RA, Dec,
                   Freq, Stokes and the desired chunking is 40 pixels in RA, 30
                   pixels in Dec, 20 pixels in Freq, and 2 pixels in Stokes,
                   chunks would be specified as [40, 30, 20, 2].
                   If dict, supported optional keys are 'l', 'm', 'frequency', 'polarization',
                   and 'time'. The supported values are positive integers,
                   indicating the length of a chunk on that particular axis. If
                   a key is missing, the associated chunk length along that axis
                   is 1. 'l' represents the longitude like dimension, and 'm'
                   represents the latitude like dimension. For apeature images,
                   'u' may be used in place of 'l', and 'v' in place of 'm'.
    :type chunks: list | dict, required
    :param mask: If specified, this is the associated image mask to read, rather than the actual
                 image pixel values. None means read the image pixel values, not the mask.
    :type mask: str, optional
    :param verbose: Emit debugging messages, default False
    :type verbose: bool
    :param blc: bottom left corner, given in the axes ordering of the input image.None=>The origin
    :type blc: a type that is convertable to a list via list(blc)
    :param trc: top right corner, given in the axes ordering of the input image.None=>image shape - 1
    :type trc: a type that is convertable to a list via list(trc)
    :return: Dask array in time, polarization, frequency, l, m order
    :rtype: dask.array

    """
    img_full_path = os.path.expanduser(infile)
    with _open_image_ro(img_full_path) as casa_image:
        """
        if isinstance(chunks, list):
            mychunks = tuple(chunks)
        elif isinstance(chunks, dict):
        """
        if isinstance(chunks, dict):
            mychunks = _get_chunk_list(
                chunks,
                casa_image.coordinates().get_names()[::-1],
                casa_image.shape()[::-1],
            )
        else:
            raise RuntimeError(
                f"incorrect type {type(chunks)} for parameter chunks. Must be "
                "either tuple or dict"
            )
        # shape list is the reverse of the actual image shape
        cshape = casa_image.shape()
        transpose_list, new_axes = _get_transpose_list(casa_image.coordinates())
        data_type = casa_image.datatype().lower()
    if data_type == "float":
        data_type = np.float32
    elif data_type == "double":
        data_type = np.float64
    elif data_type == "complex":
        data_type = np.complex64
    elif data_type == "dcomplex":
        data_type = np.complex128
    else:
        raise RuntimeError(f"Unhandled data type {data_type}")
    if verbose:
        print(f"cshape: {cshape}")
    if mask:
        img_full_path = os.sep.join([img_full_path, mask])
    # rblc and rtrc are enforced to be of type list
    if blc is None:
        rblc = len(cshape) * [0]
    else:
        if len(blc) == len(cshape):
            rblc = list(blc[::-1])
            if rblc >= cshape:
                raise RuntimeError(f"blc {blc} >= shape {cshape[::-1]}")
        else:
            raise RuntimeError(
                f"cshape has {len(cshape)} dimensions and blc has {len(blc)}"
            )
    rblc = rblc + [0 for rr in range(5) if rr >= len(mychunks)]
    if trc is None:
        rtrc = cshape
    else:
        if len(trc) == len(cshape):
            rtrc = list(trc[::-1])
            if rtrc > cshape:
                raise RuntimeError(f"trc {trc} >= shape {cshape[::-1]}")
            xblc = rblc if blc is None else list(blc)
            if list(trc) <= xblc:
                raise RuntimeError(f"trc {trc} must be > blc {blc}")
        else:
            raise RuntimeError(
                f"cshape has {len(cshape)} dimensions and blc has {len(blc)}"
            )
    rtrc = rtrc + [1 for rr in range(5) if rr >= len(mychunks)]

    # expand the actual data shape to the full 5 possible dims
    # full_shape = cshape + [1 for rr in range(5) if rr >= len(cshape)]
    # full_shape = list(rtrc - rblc)
    full_chunks = mychunks[::-1] + tuple([1 for rr in range(5) if rr >= len(mychunks)])
    d0slices = []
    for d0 in range(rblc[0], rtrc[0], full_chunks[0]):
        d0len = min(full_chunks[0], rtrc[0] - d0)
        d1slices = []

        for d1 in range(rblc[1], rtrc[1], full_chunks[1]):
            d1len = min(full_chunks[1], rtrc[1] - d1)
            d2slices = []

            for d2 in range(rblc[2], rtrc[2], full_chunks[2]):
                d2len = min(full_chunks[2], rtrc[2] - d2)
                d3slices = []

                for d3 in range(rblc[3], rtrc[3], full_chunks[3]):
                    d3len = min(full_chunks[3], rtrc[3] - d3)
                    d4slices = []

                    for d4 in range(rblc[4], rtrc[4], full_chunks[4]):
                        d4len = min(full_chunks[4], rtrc[4] - d4)

                        shapes = tuple(
                            [d0len, d1len, d2len, d3len, d4len][: len(cshape)]
                        )
                        starts = tuple([d0, d1, d2, d3, d4][: len(cshape)])
                        delayed_array = dask.delayed(_read_image_chunk)(
                            img_full_path, shapes, starts
                        )
                        d4slices += [
                            dask.array.from_delayed(delayed_array, shapes, data_type)
                        ]
                    d3slices += (
                        [dask.array.concatenate(d4slices, axis=4)]
                        if len(cshape) > 4
                        else d4slices
                    )
                d2slices += (
                    [dask.array.concatenate(d3slices, axis=3)]
                    if len(cshape) > 3
                    else d3slices
                )
            d1slices += (
                [dask.array.concatenate(d2slices, axis=2)]
                if len(cshape) > 2
                else d2slices
            )
        d0slices += (
            [dask.array.concatenate(d1slices, axis=1)] if len(cshape) > 1 else d1slices
        )
    ary = dask.array.concatenate(d0slices, axis=0)
    ary = da.expand_dims(ary, new_axes)
    return ary.transpose(transpose_list)


def _read_image_chunk(infile: str, shapes: tuple, starts: tuple) -> np.ndarray:
    with open_table_ro(infile) as tb_tool:
        data: np.ndarray = tb_tool.getcellslice(
            tb_tool.colnames()[0],
            0,
            starts,
            tuple(np.array(starts) + np.array(shapes) - 1),
        )
    return data



import toolviper.utils.logger as logger
from xradio._utils.schema import column_description_casacore_to_msv4_measure

col_to_data_variable_names = {
    "FLOAT_DATA": "SPECTRUM",
    "DATA": "VISIBILITY",
    "CORRECTED_DATA": "VISIBILITY_CORRECTED",
    "MODEL_DATA": "VISIBILITY_MODEL",
    "WEIGHT_SPECTRUM": "WEIGHT",
    "WEIGHT": "WEIGHT",
    "FLAG": "FLAG",
    "UVW": "UVW",
    "TIME_CENTROID": "TIME_CENTROID",
    "EXPOSURE": "EFFECTIVE_INTEGRATION_TIME",
}
col_dims = {
    "DATA": ("time", "baseline_id", "frequency", "polarization"),
    "CORRECTED_DATA": ("time", "baseline_id", "frequency", "polarization"),
    "MODEL_DATA": ("time", "baseline_id", "frequency", "polarization"),
    "WEIGHT_SPECTRUM": ("time", "baseline_id", "frequency", "polarization"),
    "WEIGHT": ("time", "baseline_id", "frequency", "polarization"),
    "FLAG": ("time", "baseline_id", "frequency", "polarization"),
    "UVW": ("time", "baseline_id", "uvw_label"),
    "TIME_CENTROID": ("time", "baseline_id"),
    "EXPOSURE": ("time", "baseline_id"),
    "FLOAT_DATA": ("time", "baseline_id", "frequency", "polarization"),
}
col_to_coord_names = {
    "TIME": "time",
    "ANTENNA1": "baseline_ant1_id",
    "ANTENNA2": "baseline_ant2_id",
}


def create_attribute_metadata(col, main_column_descriptions):
    attrs_metadata = column_description_casacore_to_msv4_measure(
        main_column_descriptions[col]
    )
    if col in ["DATA", "CORRECTED_DATA", "WEIGHT"]:
        if not attrs_metadata:
            attrs_metadata["type"] = "quanta"
            attrs_metadata["units"] = ["unkown"]

    return attrs_metadata



import itertools
import toolviper.utils.logger as logger
from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np
import xarray as xr

from casacore import tables

from ._tables.table_query import open_table_ro, open_query
from ._tables.read import table_exists


def enumerated_product(*args):
    yield from zip(
        itertools.product(*(range(len(x)) for x in args)), itertools.product(*args)
    )


def create_partitions(in_file: str, partition_scheme: list) -> list[dict]:
    """Create a list of dictionaries with the partition information.

    Parameters
    ----------
    in_file: str
        Input MSv2 file path.
    partition_scheme:  list
        A MS v4 can only contain a single data description (spectral window and polarization setup), and observation mode. Consequently, the MS v2 is partitioned when converting to MS v4.
        In addition to data description and polarization setup a finer partitioning is possible by specifying a list of partitioning keys. Any combination of the following keys are possible:
        "FIELD_ID", "SCAN_NUMBER", "STATE_ID", "SOURCE_ID", "SUB_SCAN_NUMBER".
        For mosaics where the phase center is rapidly changing (such as VLA on the fly mosaics)  partition_scheme should be set to an empty list []. By default, ["FIELD_ID"].
    Returns
    -------
    list
        list of dictionaries with the partition information.
    """
    # vla_otf (bool, optional):  The partioning of VLA OTF (on the fly) mosaics needs a special partitioning scheme. Defaults to False.

    # Create partition table
    import pandas as pd
    import os

    partition_scheme = ["DATA_DESC_ID", "OBS_MODE", "OBSERVATION_ID"] + partition_scheme

    # Open MSv2 tables and add columns to partition table (par_df):
    par_df = pd.DataFrame()
    main_tb = tables.table(
        in_file, readonly=True, lockoptions={"option": "usernoread"}, ack=False
    )
    par_df["DATA_DESC_ID"] = main_tb.getcol("DATA_DESC_ID")
    par_df["FIELD_ID"] = main_tb.getcol("FIELD_ID")
    par_df["SCAN_NUMBER"] = main_tb.getcol("SCAN_NUMBER")
    par_df["STATE_ID"] = main_tb.getcol("STATE_ID")
    par_df["OBSERVATION_ID"] = main_tb.getcol("OBSERVATION_ID")
    par_df["ANTENNA1"] = main_tb.getcol("ANTENNA1")
    par_df = par_df.drop_duplicates()

    field_tb = tables.table(
        os.path.join(in_file, "FIELD"),
        readonly=True,
        lockoptions={"option": "usernoread"},
        ack=False,
    )
    # if vla_otf:
    #     par_df["FIELD_NAME"] = np.array(field_tb.getcol("NAME"))[par_df["FIELD_ID"]]

    # Get source ids if available from source table.
    if table_exists(os.path.join(os.path.join(in_file, "SOURCE"))):
        source_tb = tables.table(
            os.path.join(in_file, "SOURCE"),
            readonly=True,
            lockoptions={"option": "usernoread"},
            ack=False,
        )
        if source_tb.nrows() != 0:
            par_df["SOURCE_ID"] = field_tb.getcol("SOURCE_ID")[par_df["FIELD_ID"]]
            # if vla_otf:
            #     par_df["SOURCE_NAME"] = np.array(source_tb.getcol("NAME"))[
            #         par_df["SOURCE_ID"]
            #     ]

    # Get intents and subscan numbers if available from state table.
    if table_exists(os.path.join(in_file, "STATE")):
        state_tb = tables.table(
            os.path.join(in_file, "STATE"),
            readonly=True,
            lockoptions={"option": "usernoread"},
            ack=False,
        )
        if state_tb.nrows() != 0:
            # print('state_tb',state_tb.nrows(),state_tb)
            par_df["OBS_MODE"] = np.array(state_tb.getcol("OBS_MODE"))[
                par_df["STATE_ID"]
            ]
            par_df["SUB_SCAN_NUMBER"] = state_tb.getcol("SUB_SCAN")[par_df["STATE_ID"]]
        else:
            par_df.drop(["STATE_ID"], axis=1)

    # Check if all partition scheme criteria are present in the partition table.
    partition_scheme_updated = []
    partition_criteria = {}
    for par in partition_scheme:
        if par in par_df.columns:
            partition_criteria[par] = par_df[par].unique()
            partition_scheme_updated.append(par)
    logger.info(f"Partition scheme that will be used: {partition_scheme_updated}")

    # Make all possible combinations of the partition criteria.
    enumerated_partitions = enumerated_product(*list(partition_criteria.values()))

    # print('par_df',par_df)

    # Create a list of dictionaries with the partition information. This will be used to query the MSv2 main table.
    partitions = []
    partition_axis_names = [
        "DATA_DESC_ID",
        "OBSERVATION_ID",
        "FIELD_ID",
        "SCAN_NUMBER",
        "STATE_ID",
        "SOURCE_ID",
        "OBS_MODE",
        "SUB_SCAN_NUMBER",
    ]
    if "ANTENNA1" in partition_scheme:
        partition_axis_names.append("ANTENNA1")

    for idx, pair in enumerated_partitions:
        query = ""
        for i, par in enumerate(partition_scheme_updated):
            if isinstance(pair[i], str):
                query = query + f'{par} == "{pair[i]}" and '
            else:
                query = query + f"{par} == {pair[i]} and "
        query = query[:-4]  # remove last and
        sub_par_df = par_df.query(query).drop_duplicates()

        if sub_par_df.shape[0] != 0:
            partition_info = {}

            # FIELD_NAME	SOURCE_NAME
            for col_name in partition_axis_names:
                if col_name in sub_par_df.columns:
                    partition_info[col_name] = sub_par_df[col_name].unique()
                else:
                    partition_info[col_name] = [None]

            partitions.append(partition_info)

    return partitions


# Used by code that will be deprecated at some stage. See #192
# Still need to clarify what to do about intent string filtering ('WVR', etc.)


def make_partition_ids_by_ddi_intent(
    infile: str, spw_names: xr.DataArray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Produces arrays of per-partition ddi, scan, state_id, for when
    using the partition scheme 'intents' (ddi, scan, subscans(state_ids))

    Parameters
    ----------
    infile : str
        return: arrays with indices that define every partition
    spw_names: xr.DataArray


    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        arrays with indices that define every partition
    """
    # TODO: could explore other TAQL alternatives, like
    # select ... from ::STATE where OBS_MODE = ...
    #
    # This will work only if intents are already alphabetically sorted (grouped),
    # won't work for alternating intents:
    # taql_intents = "select rowid() as ROWS from $state_tbl GROUPBY OBS_MODE "

    with open_table_ro(str(Path(infile, "STATE"))) as state_tbl:
        distinct_obs_mode = find_distinct_obs_mode(infile, state_tbl)

        if distinct_obs_mode is None:
            return partition_when_empty_state(infile)

        with open_table_ro(infile) as main_tbl:
            (
                data_desc_id,
                state_id_partitions,
                intent_names,
            ) = make_ddi_state_intent_lists(
                main_tbl, state_tbl, distinct_obs_mode, spw_names
            )

    # Take whatever scans given by the STATE_IDs and DDIs
    scan_number = [None] * len(state_id_partitions)

    return data_desc_id, scan_number, state_id_partitions, intent_names


def find_distinct_obs_mode(
    infile: str, state_table: tables.table
) -> Union[List[str], None]:
    """
    Produce a list of unique "scan/subscan" intents.

    Parameters
    ----------
    infile : str
        Path to the MS
    state_table : tables.table
        casacore table object to read from

    Returns
    -------
    Union[List[str], None]
        List of unique "scan/subscan" intents as given in the
        OBS_MODE column of the STATE subtable. None if the STATE subtable
        is empty or there is a problem reading it
    """
    taql_distinct_intents = "select DISTINCT OBS_MODE from $state_table"
    with open_query(state_table, taql_distinct_intents) as query_intents:
        if query_intents.nrows() == 0:
            logger.warning(
                "STATE subtable has no data. Cannot partition by scan/subscan intent"
            )
            return None

        distinct_obs_mode = query_intents.getcol("OBS_MODE")
        logger.debug(
            f"  Query for distinct OBS_MODE len: {len(distinct_obs_mode)}, values: {distinct_obs_mode}"
        )
        return distinct_obs_mode


def make_ddi_state_intent_lists(
    main_tbl: tables.table,
    state_tbl: tables.table,
    distinct_obs_mode: np.ndarray,
    spw_name_by_ddi: Dict[int, str],
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Produce arrays of (ddi indices, state indices, intent string)
    for every distinct intent string, where every item represents one
    partition of the main table

    As the obs_mode strings have concatenated intent strings from all
    the scan and subscan intents, this function has started
    implementing some simple heuristics to remove the intent items
    that are not related to the respective DDIs (for example WVR
    intent is the only kept when the DDI/SPW has WVR in its name). See
    call to filter_intents_per_ddi()

    Parameters
    ----------
    main_tbl : tables.table
        main MS table openend as a casacore.tables.table
    state_tbl : tables.table
        STATE subtable openend as a casacore.tables.table
    distinct_obs_mode : np.ndarray
        list of unique/distinct OBS_MODE strings from the STATE table
    spw_name_by_ddi: Dict[int, str]


    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]
        arrays of (ddi indices, state indices, intent string)
    """
    data_desc_id, state_id_partitions, intent_names = [], [], []
    for intent in distinct_obs_mode:
        where_intent = f"where OBS_MODE = '{intent}'"
        taql_states = f"select ROWID() as ROWS from $state_tbl {where_intent}"
        with open_query(state_tbl, taql_states) as query_states:
            state_ids_for_intent = query_states.getcol("ROWS")

        state_ids = " OR STATE_ID = ".join(np.char.mod("%d", state_ids_for_intent))
        taql_partition = (
            f"select DISTINCT DATA_DESC_ID from $main_tbl where STATE_ID = {state_ids}"
        )
        with open_query(main_tbl, taql_partition) as query_ddi_intent:
            # No data for these STATE_IDs
            if query_ddi_intent.nrows() == 0:
                continue

            # Will implicitly take whatever scans given the STATE_IDs
            # and DDIs scan_number. Not needed:
            # scan_number = query_ddi_intent.getcol("SCAN_NUMBER")
            ddis = query_ddi_intent.getcol("DATA_DESC_ID")

            data_desc_id.extend(ddis)
            state_id_partitions.extend([state_ids_for_intent] * len(ddis))

            # Try to select/exclude confusing or mixed intent names such as 'WVR#*'
            intents_ddi = filter_intents_per_ddi(ddis, "WVR", intent, spw_name_by_ddi)
            intent_names.extend(intents_ddi)

    logger.debug(
        f"Produced data_desc_id: {data_desc_id},\n state_id_partitions: {state_id_partitions}"
    )
    return data_desc_id, state_id_partitions, intent_names


def filter_intents_per_ddi(
    ddis: List[int], substr: str, intents: str, spw_name_by_ddi: Dict[int, str]
) -> List[str]:
    """
    For a given pair of:
    - substring (say 'WVR') associated with a type of intent we want to differentiate
    - intents string (multiple comma-separated scan/subscan intents)
    => do: for every DDI passed in the list of ddis, either keep only the
    intents that have that substring (if there are any) or drop them, depending on
    whether that substring is present in the SPW name. This is to filter in only
    the intents that really apply to every DDI/SPW.

    Parameters
    ----------
    ddis : List[int]
        list of ddis for which the intents have to be filtered
    substr : str
        substring to filter by
    intents : str
        string with a comma-separated list of individual
        scan/subscan intent strings (like scan/subscan intents as stored
        in the MS STATE/OBS_MODE
    spw_name_by_ddi : Dict[int, str]
        SPW names by DDI ID (row index) key

    Returns
    -------
    List[str]
        list where the intents related to 'substr' have been filtered in our out
    """
    present = substr in intents
    # Nothing to effectively filter, full cs-list of intents apply to all DDIs
    if not present:
        return [intents] * len(ddis)

    every_intent = intents.split(",")
    filtered_intents = []
    for ddi in ddis:
        spw_name = spw_name_by_ddi.get(ddi, "")

        if not spw_name:
            # we cannot say / cannot filter
            filtered_intents.append(intents)
            continue

        # A not-xor to select/deselect (or keep-only/drop) the intents that apply
        # to this DDI
        ddi_intents = [
            intnt for intnt in every_intent if (substr in intnt) == (substr in spw_name)
        ]
        ddi_intents = ",".join(ddi_intents)
        filtered_intents.append(ddi_intents)

    return filtered_intents


def partition_when_empty_state(
    infile: str,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Generate fallback partition ids when trying to partition by
    'intent' but the STATE table is empty.

    Some MSs have no STATE rows and in the main table STATE_ID==-1
    (that is not a valid MSv2 but it happens).

    Parameters
    ----------
    infile : str
        Path to the MS

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        same as make_partition_ids_by_ddi_intent but with
        effectively only ddi indices and other indices set to None ("any
        IDs found")
    """
    try:
        main_table = None

        main_table = tables.table(
            infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
        )
        taql_ddis = "select DISTINCT DATA_DESC_ID from $main_table"
        with open_query(main_table, taql_ddis) as query_per_intent:
            # Will take whatever scans given the STATE_IDs and DDIs
            # scan_number = query_per_intent.getcol("SCAN_NUMBER")
            distinct_ddis = query_per_intent.getcol("DATA_DESC_ID")

        logger.debug(
            f"Producing {len(distinct_ddis)} partitions for ddis: {distinct_ddis}"
        )
        nparts = len(distinct_ddis)

    finally:
        if main_table:
            main_table.close()

    return distinct_ddis, [None] * nparts, [None] * nparts, [""] * nparts



import os, time
from typing import List, Optional, Union

import dask
import numpy as np
import xarray as xr

from ..._utils.xds_helper import flatten_xds, calc_optimal_ms_chunk_shape
from .write import write_generic_table, write_main_table_slice
from .write import create_table, revert_time

from casacore import tables


# TODO: this should be consolidated with the equivalent in read_main_table,
# if we keep this mapping
rename_to_msv2_cols = {
    "antenna1_id": "ANTENNA1",
    "antenna2_id": "ANTENNA2",
    "feed1_id": "FEED1",
    "feed2_id": "FEED2",
    # optional cols:
    # "WEIGHT": "WEIGHT_SPECTRUM",
    "VIS_CORRECTED": "CORRECTED_DATA",
    "VIS": "DATA",
    "VIS_MODEL": "MODEL_DATA",
    "AUTOCORR": "FLOAT_DATA",
}
# cols added in xds not in MSv2
cols_not_in_msv2 = ["baseline_ant1_id", "baseline_ant2_id"]


def cols_from_xds_to_ms(cols: List[str]) -> List[str]:
    """
    Translates between lowercase/uppercase convention
    Rename some MS_colum_names <-> xds_data_var_names
    Excludes the pointing_ vars that are in the xds but should not be written to MS
    """
    return {
        rename_to_msv2_cols.get(col, col).upper(): col
        for col in cols
        if (col and col not in cols_not_in_msv2 and not col.startswith("pointing_"))
    }


def write_ms(
    mxds: xr.Dataset,
    outfile: str,
    infile: str = None,
    subtables: bool = False,
    modcols: Union[List[str], None] = None,
    verbose: bool = False,
    execute: bool = True,
) -> Optional[list]:
    """
    Write ms format xds contents back to casacore MS (CTDS - casacore Table Data System) format on disk

    Parameters
    ----------
    mxds : xr.Dataset,
        Source multi-xarray dataset (originally created by read_ms)
    outfile : str
        Destination filename
    infile : Union[str, None] (Default value = None)
        Source filename to copy subtables from. Generally faster than reading/writing through mxds via the subtables parameter. Default None
        does not copy subtables to output.
    subtables : bool (Default value = False)
        Also write subtables from mxds. Default of False only writes mxds attributes that begin with xdsN to the MS main table.
        Setting to True will write all other mxds attributes to subtables of the main table.  This is probably going to be SLOW!
        Use infile instead whenever possible.
    modcols : Union[List[str], None] (Default value = None)
        List of strings indicating what column(s) were modified (aka xds data_vars). Different logic can be applied to speed up processing when
        a data_var has not been modified from the input. Default None assumes everything has been modified (SLOW)
    verbose : bool (Default value = False)
        Whether or not to print output progress. Since writes will typically execute the DAG, if something is
        going to go wrong, it will be here.  Default False
    execute : bool (Default value = True)
        Whether or not to actually execute the DAG, or just return it with write steps appended. Default True will execute it

    Returns
    -------
    Optional[list]
        delayed write functions
    """
    outfile = os.path.expanduser(outfile)
    if verbose:
        print("initializing output...")
    start = time.time()

    xds_list = [flatten_xds(xds) for _key, xds in mxds.partitions.items()]

    cols = cols_from_xds_to_ms(
        list(set([dv for dx in xds_list for dv in dx.data_vars]))
    )
    if modcols is None:
        modcols = cols

    # create an empty main table with enough space for all desired xds partitions
    # the first selected xds partition will be passed to create_table to provide a definition of columns and table keywords
    # we first need to add in additional keywords for the selected subtables that will be written as well
    max_rows = np.sum([dx.row.shape[0] for dx in xds_list])
    create_table(
        outfile, xds_list[0], max_rows=max_rows, infile=infile, cols=cols, generic=False
    )

    # start a list of dask delayed writes to disk (to be executed later)
    # the SPECTRAL_WINDOW, POLARIZATION, and DATA_DESCRIPTION tables must always be present and will always be written
    delayed_writes = [
        dask.delayed(write_generic_table)(
            mxds.metainfo["spectral_window"], outfile, "SPECTRAL_WINDOW", cols=None
        )
    ]
    delayed_writes += [
        dask.delayed(write_generic_table)(
            mxds.metainfo["polarization"], outfile, "POLARIZATION", cols=None
        )
    ]
    # should data_description be kept somewhere (in attrs?) or rebuilt?
    # delayed_writes += [
    #     dask.delayed(write_generic_table)(
    #         mxds.metainfo["data_description"], outfile, "DATA_DESCRIPTION", cols=None
    #     )
    # ]
    if subtables:  # also write the rest of the subtables
        for subtable in list(mxds.attrs.keys()):
            if (
                subtable.startswith("xds")
                or (subtable in ["spectral_window", "polarization", "data_description"])
                or not isinstance(subtable, xr.Dataset)
            ):
                continue

            if verbose:
                print("writing subtable %s..." % subtable)
            delayed_writes += [
                dask.delayed(write_generic_table)(
                    mxds.attrs[subtable], outfile, subtable, cols=None
                )
            ]

    ddi_row_start = 0  # output rows will be ordered by DDI
    for xds in xds_list:
        txds = xds.copy().unify_chunks()
        # TODO: carry over or rebuild?
        ddi = 0  # txds.data_desc_id[:1].values[0]

        # serial write entire DDI column first so subsequent delayed writes can find their spot
        if verbose:
            print("setting up DDI %i..." % ddi)

        # write each chunk of each modified data_var, triggering the DAG along the way
        for col in modcols:
            if col not in txds:
                continue  # this can happen with bad_cols, should still be created in create_table()

            if col in cols_not_in_msv2:
                continue

            chunks = txds[col].chunks
            dims = txds[col].dims
            for d0 in range(len(chunks[0])):
                d0start = ([0] + list(np.cumsum(chunks[0][:-1])))[d0]

                for d1 in range(len(chunks[1]) if len(chunks) > 1 else 1):
                    d1start = (
                        ([0] + list(np.cumsum(chunks[1][:-1])))[d1]
                        if len(chunks) > 1
                        else 0
                    )

                    for d2 in range(len(chunks[2]) if len(chunks) > 2 else 1):
                        d2start = (
                            ([0] + list(np.cumsum(chunks[2][:-1])))[d2]
                            if len(chunks) > 2
                            else 0
                        )

                        starts = [d0start, d1start, d2start]
                        lengths = [
                            chunks[0][d0],
                            (chunks[1][d1] if len(chunks) > 1 else 0),
                            (chunks[2][d2] if len(chunks) > 2 else 0),
                        ]
                        slices = [
                            slice(starts[0], starts[0] + lengths[0]),
                            slice(starts[1], starts[1] + lengths[1]),
                            slice(starts[2], starts[2] + lengths[2]),
                        ]
                        txda = txds[col].isel(
                            dict(zip(dims, slices)), missing_dims="ignore"
                        )
                        starts[0] = starts[0] + ddi_row_start  # offset to end of table
                        delayed_writes += [
                            dask.delayed(write_main_table_slice)(
                                txda,
                                outfile,
                                ddi=ddi,
                                col=col,
                                full_shape=txds[col].shape[1:],
                                starts=starts,
                            )
                        ]

        # now write remaining data_vars from the xds that weren't modified
        # this can be done faster by collapsing the chunking to maximum size (minimum #) possible
        max_chunk_size = np.prod(
            [txds.chunks[kk][0] for kk in txds.chunks if kk in ["row", "freq", "pol"]]
        )
        for col in list(np.setdiff1d(list(cols), modcols)):
            if col not in txds:
                continue  # this can happen with bad_cols, should still be created in create_table()

            if col in cols_not_in_msv2:
                continue

            col_chunk_size = np.prod([kk[0] for kk in txds[col].chunks])
            if max_chunk_size <= 0:
                max_chunk_size = 19200
            if col_chunk_size <= 0:
                col_rows = max_chunk_size
            else:
                col_rows = (
                    int(np.ceil(max_chunk_size / col_chunk_size))
                    * txds[col].chunks[0][0]
                )
            for rr in range(0, txds[col].row.shape[0], col_rows):
                txda = txds[col].isel(row=slice(rr, rr + col_rows))
                delayed_writes += [
                    dask.delayed(write_main_table_slice)(
                        txda,
                        outfile,
                        ddi=ddi,
                        col=rename_to_msv2_cols.get(col, col).upper(),
                        full_shape=txda.shape[1:],
                        starts=(rr + ddi_row_start,) + (0,) * (len(txda.shape) - 1),
                    )
                ]

        ddi_row_start += txds.row.shape[0]  # next xds will be appended after this one

    if execute:
        if verbose:
            print("triggering DAG...")
        zs = dask.compute(delayed_writes)
        if verbose:
            print(
                "execution time %0.2f sec. Compute result len: %d"
                % ((time.time() - start), len(zs))
            )
    else:
        if verbose:
            print("returning delayed task list")
        return delayed_writes


def write_ms_serial(
    mxds: xr.Dataset,
    outfile: str,
    infile: str = None,
    subtables: bool = False,
    verbose: bool = False,
    execute: bool = True,
    memory_available_in_bytes: int = 500000000000,
):
    """
    Write ms format xds contents back to casacore table format on disk

    Parameters
    ----------
    mxds : xr.Dataset
        Source multi-xarray dataset (originally created by read_ms)
    outfile : str
        Destination filename
    infile : str (Default value = None)
        Source filename to copy subtables from. Generally faster than reading/writing through mxds via the subtables parameter. Default None
        does not copy subtables to output.
    subtables : bool (Default value = False)
        Also write subtables from mxds. Default of False only writes mxds attributes that begin with xdsN to the MS main table.
        Setting to True will write all other mxds attributes to subtables of the main table.  This is probably going to be SLOW!
        Use infile instead whenever possible.
    verbose : bool (Default value = False)
        Whether or not to print output progress. Since writes will typically execute the DAG, if something is
        going to go wrong, it will be here.  Default False

    execute : bool (Default value = True)
        Whether or not to actually execute the DAG, or just return it with write steps appended. Default True will execute it
    memory_available_in_bytes : (Default value = 500000000000)

    Returns
    -------

    """

    print("*********************")
    outfile = os.path.expanduser(outfile)
    if verbose:
        print("initializing output...")
    # start = time.time()

    xds_list = [flatten_xds(xds) for _key, xds in mxds.partitions.items()]
    cols = list(set([dv for dx in xds_list for dv in dx.data_vars]))
    cols = cols_from_xds_to_ms(list(np.atleast_1d(cols)))

    # create an empty main table with enough space for all desired xds partitions
    # the first selected xds partition will be passed to create_table to provide a definition of columns and table keywords
    # we first need to add in additional keywords for the selected subtables that will be written as well
    max_rows = np.sum([dx.row.shape[0] for dx in xds_list])
    create_table(
        outfile, xds_list[0], max_rows=max_rows, infile=infile, cols=cols, generic=False
    )

    # start a list of dask delayed writes to disk (to be executed later)
    # the SPECTRAL_WINDOW, POLARIZATION, and DATA_DESCRIPTION tables must always be present and will always be written
    write_generic_table(
        mxds.metainfo["spectral_window"], outfile, "SPECTRAL_WINDOW", cols=None
    )
    write_generic_table(
        mxds.metainfo["polarization"], outfile, "POLARIZATION", cols=None
    )
    # should data_description be kept somewhere (in attrs?) or rebuilt?
    # write_generic_table(mxds.metainfo.data_description, outfile, "DATA_DESCRIPTION", cols=None)

    if subtables:  # also write the rest of the subtables
        # for subtable in list(mxds.attrs.keys()):
        #'OBSERVATION','STATE'
        # ['FEED','OBSERVATION','FIELD','ANTENNA','HISTORY','STATE']
        # ['FEED','FIELD','ANTENNA','HISTORY']
        # ,'FIELD','ANTENNA'
        # for subtable in ['OBSERVATION']:
        for subtable in list(mxds.metainfo.keys()):
            if subtable.startswith("xds") or (
                subtable in ["spectral_window", "polarization", "data_description"]
            ):
                continue
            if verbose:
                print("writing subtable %s..." % subtable)
            # print(subtable)
            # print(mxds.attrs[subtable])
            try:
                write_generic_table(
                    mxds.metainfo[subtable], outfile, subtable.upper(), cols=None
                )
            except (RuntimeError, KeyError) as exc:
                print(f"Exception writing subtable {subtable}: {exc}")

    part_key0 = next(iter(mxds.partitions))
    vis_data_shape = mxds.partitions[part_key0].VIS.shape
    rows_chunk_size = calc_optimal_ms_chunk_shape(
        memory_available_in_bytes, vis_data_shape, 16, "DATA"
    )

    # print(rows_chunk_size)
    # rows_chunk_size = 200000000
    # write each chunk of each modified data_var, triggering the DAG along the way
    tbs = tables.table(
        outfile, readonly=False, lockoptions={"option": "permanentwait"}, ack=True
    )

    start_main = time.time()
    for col, var_name in cols.items():
        xda = mxds.partitions[part_key0][var_name]
        # print(col,xda.dtype)

        for start_row in np.arange(0, vis_data_shape[0], rows_chunk_size):
            end_row = start_row + rows_chunk_size
            if end_row > vis_data_shape[0]:
                end_row = vis_data_shape[0]

            # start = time.time()
            values = xda[start_row:end_row,].compute().values
            if xda.dtype == "datetime64[ns]":
                values = revert_time(values)
            # print('1. Time', time.time()-start, values.shape)

            # start = time.time()
            try:
                tbs.putcol(col, values, start_row, len(values))
                # print('2. Time', time.time()-start)
            except RuntimeError as exc:
                print(f"Exception writing main table column {col}: {exc}")

    print("3. Time", time.time() - start_main)

    tbs.unlock()
    tbs.close()



from dataclasses import dataclass
import typing

__all__ = [
    "AttrSchemaRef",
    "ArraySchema",
    "ArraySchemaRef",
    "DatasetSchema",
    "DictSchema",
]


@dataclass(frozen=True)
class AttrSchemaRef:
    """
    Schema information about an attribute as referenced from an array or
    dataset schema.

    This includes the name and docstring associated with the attribute
    in the array or dataset schema definition.
    """

    name: str
    """Name of attribute as given in data array / dataset."""
    typ: type
    """
    Python type of attribute. Note that this might again be a data
    array or dataset, but we don't track that explicitly.
    """
    optional: bool
    """Is the attribute optional?"""
    default: typing.Optional[typing.Any]
    """If optional: What is the default value?"""
    docstring: str
    """Documentation string of attribute reference"""


@dataclass(frozen=True)
class ArraySchema:
    """
    Schema for xarray data array

    A data array maps a tuple of dimensions to (numpy) values. The schema
    allows for multiple options both on dimensions as well as types to be
    used.
    """

    schema_name: str
    """(Class) name of the schema"""
    dimensions: typing.List[typing.List[str]]
    """List of possible dimensions"""
    dtypes: typing.List[typing.List["numpy.dtype"]]
    """List of possible (numpy) types"""

    coordinates: typing.List["ArraySchemaRef"]
    """Coordinates data arrays giving values to dimensions"""
    attributes: typing.List[AttrSchemaRef]
    """Attributes associated with data array"""

    class_docstring: typing.Optional[str]
    """Documentation string of class"""
    data_docstring: typing.Optional[str]
    """Documentation string of data in class"""

    def is_coord(self) -> bool:
        """
        Checks with this is a valid coordinate data array

        Such data arrays must not have coordinate references of their own,
        i.e. be defined in terms of (integer) dimensions only. This is of
        course because their very purpose is to map these integer dimensions
        to semantically meaningful values, such as frequencies.
        """
        return not self.coordinates

    def required_dimensions(self) -> [str]:
        """
        Returns set of dimensions that is always required
        """

        req_dims = set(self.dimensions[0])
        for dims in self.dimensions[1:]:
            req_dims &= set(dims)
        return req_dims


@dataclass(frozen=True)
class ArraySchemaRef(ArraySchema):
    """
    Schema for xarray data array as referenced from a dataset schema

    Includes information about name and docstring associated with
    array schema where referenced
    """

    name: str
    """Name of array schema as given in dataset."""
    optional: bool
    """Is the data array optional?"""
    default: typing.Optional[typing.Any]
    """If optional: What is the default value?"""
    docstring: typing.Optional[str] = None
    """Documentation string of array reference"""


@dataclass(frozen=True)
class DatasetSchema:
    """
    Schema for an xarray dataset
    """

    schema_name: str
    """(Class) name of the schema"""

    dimensions: [[str]]
    """List of possible dimensions (derived from data arrays)"""
    coordinates: [ArraySchemaRef]
    """List of coordinate data arrays"""
    data_vars: [ArraySchemaRef]
    """List of data arrays"""
    attributes: [AttrSchemaRef]
    """List of attributes"""

    class_docstring: typing.Optional[str]
    """Documentation string of class"""


@dataclass(frozen=True)
class DictSchema:
    """
    Schema for a simple dictionary
    """

    schema_name: str
    """(Class) name of the schema"""

    attributes: [AttrSchemaRef]
    """List of attributes"""

    class_docstring: typing.Optional[str]
    """Documentation string of class"""



import os

import toolviper.utils.logger as logger

from pathlib import Path
from typing import Dict, List

import xarray as xr

from ._tables.read import load_generic_table, table_exists
from ._tables.read_subtables import read_ephemerides, read_delayed_pointing_table


subt_rename_ids = {
    "ANTENNA": {"row": "antenna_id", "dim_1": "xyz"},
    "FEED": {"dim_1": "xyz", "dim_2": "receptor", "dim_3": "receptor2"},
    "FIELD": {"row": "field_id", "dim_1": "poly_id", "dim_2": "ra/dec"},
    "FREQ_OFFSET": {"antenna1": "antenna1_id", "antenna2": "antenna2_id"},
    "OBSERVATION": {"row": "observation_id", "dim_1": "start/end"},
    "POINTING": {"dim_1": "n_polynomial", "dim_3": "dir"},
    "POLARIZATION": {"row": "pol_setup_id", "dim_2": "product_id"},
    "PROCESSOR": {"row": "processor_id"},
    "SPECTRAL_WINDOW": {"row": "spectral_window_id", "dim_1": "chan"},
    "SOURCE": {"dim_1": "ra/dec", "dim_2": "line"},
    "STATE": {"row": "state_id"},
    "SYSCAL": {"dim_1": "frequency", "dim_2": "receptor"},
    # Would make sense for non-std "WS_NX_STATION_POSITION"
    "WEATHER": {"dim_1": "xyz"},
}


def read_ms_subtables(
    infile: str, done_subt: List[str], asdm_subtables: bool = False
) -> Dict[str, xr.Dataset]:
    """
    Read MSv2 subtables (main table keywords) as xr.Dataset

    Parameters
    ----------
    infile : str
        input MeasurementSet path
    done_subt : List[str]
        Subtables that were already read, to skip them
    asdm_subtables : bool (Default value = False)
        Whether to also read ASDM_* subtables

    Returns
    -------
    Dict[str, xr.Dataset]
        dict of xarray datasets read from subtables (metadata tables)

    """
    ignore_msv2_cols_subt = ["FLAG_CMD", "FLAG_ROW", "BEAM_ID"]
    skip_tables = ["SORTED_TABLE", "FLAG_CMD"] + done_subt
    stbl_list = sorted(
        [
            tname
            for tname in os.listdir(infile)
            if (tname not in skip_tables)
            and (os.path.isdir(os.path.join(infile, tname)))
            and (table_exists(os.path.join(infile, tname)))
        ]
    )

    subtables = {}
    for _ii, subt_name in enumerate(stbl_list):
        if not asdm_subtables and subt_name.startswith("ASDM_"):
            logger.debug(f"skipping ASDM_ subtable {subt_name}...")
            continue
        else:
            logger.debug(f"reading subtable {subt_name}...")

        if subt_name == "POINTING":
            subt_path = Path(infile, subt_name)
            rename_ids = {"dim_2": "n_polynomial", "dim_3": "dir"}
            xds = read_delayed_pointing_table(str(subt_path), rename_ids=rename_ids)
        else:
            xds = load_generic_table(
                infile,
                subt_name,
                timecols=["TIME"],
                ignore=ignore_msv2_cols_subt,
                rename_ids=subt_rename_ids.get(subt_name, None),
            )

        if len(xds.sizes) != 0:
            subtables[subt_name.lower()] = xds

    if "field" in subtables:
        ephem_xds = read_ephemerides(infile)
        if ephem_xds:
            subtables["ephemerides"] = ephem_xds

    return subtables


def add_pointing_to_partition(
    xds_part: xr.Dataset, xds_pointing: xr.Dataset
) -> xr.Dataset:
    """
    Take pointing variables from a (delayed) pointing dataset and
    transfer them to a main table partition dataset (interpolating into
    the destination time axis)

    Parameters
    ----------
    xds_part : xr.Dataset
        a partition/sub-xds of the main table
    xds_pointing : xr.Dataset
        the xds read from the pointing subtable

    Returns
    -------
    xr.Dataset
        partition xds with pointing variables added/interpolated from the
        pointing_xds into its time axis

    """
    interp_xds = xds_pointing.interp(time=xds_part.time, method="nearest")
    for var in interp_xds.data_vars:
        xds_part[f"pointing_{var}"] = interp_xds[var]

    return xds_part



import os

from xradio.measurement_set import ProcessingSet
import toolviper.utils.logger as logger
from xradio._utils.zarr.common import _open_dataset, _get_file_system_and_items
import s3fs


def open_processing_set(
    ps_store: str,
    intents: list = None,
) -> ProcessingSet:
    """Creates a lazy representation of a Processing Set (only meta-data is loaded into memory).

    Parameters
    ----------
    ps_store : str
        String of the path and name of the processing set. For example '/users/user_1/uid___A002_Xf07bba_Xbe5c_target.lsrk.vis.zarr'.
    intents : list, optional
        A list of intents to be open for example ['OBSERVE_TARGET#ON_SOURCE']. The intents in a processing set can be seen by calling processing_set.summary().
        By default None, which will open all intents.

    Returns
    -------
    processing_set
        Lazy representation of processing set (data is represented by Dask.arrays).
    """
    from xradio.measurement_set import MeasurementSetXds

    file_system, ms_store_list = _get_file_system_and_items(ps_store)

    ps = ProcessingSet()
    data_group = "base"
    for ms_name in ms_store_list:
        # try:
        ms_store = os.path.join(ps_store, ms_name)
        correlated_store = os.path.join(ms_store, "correlated_xds")

        xds = _open_dataset(correlated_store, file_system)
        data_groups = xds.attrs["data_groups"]

        if (intents is None) or (
            bool(set(xds.attrs["partition_info"]["intents"]).intersection(intents))
        ):
            sub_xds_dict, field_and_source_xds_dict = _open_sub_xds(
                ms_store, file_system=file_system, data_groups=data_groups
            )

            xds.attrs = {
                **xds.attrs,
                **sub_xds_dict,
            }

            for data_group_name, data_group_vals in data_groups.items():
                xds[data_group_vals["correlated_data"]].attrs[
                    "field_and_source_xds"
                ] = field_and_source_xds_dict[data_group_name]

            ps[ms_name] = MeasurementSetXds(xds)
        # except Exception as e:
        #     logger.warning(f"Could not open {ms_name} due to {e}")
        #     continue

    return ps


def _open_sub_xds(ms_store, file_system, data_groups, load=False):
    sub_xds_dict = {}
    field_and_source_xds_dict = {}

    if isinstance(file_system, s3fs.core.S3FileSystem):
        file_names = [
            bd.split(sep="/")[-1] for bd in file_system.listdir(ms_store, detail=False)
        ]
    else:
        file_names = file_system.listdir(ms_store)
    file_names = [item for item in file_names if not item.startswith(".")]

    file_names.remove("correlated_xds")

    field_dict = {"field_and_source_xds_" + key: key for key in data_groups.keys()}

    # field_and_source_xds_name_start = "FIELD"
    for n in file_names:
        xds = _open_dataset(
            os.path.join(ms_store, n), load=load, file_system=file_system
        )
        # Skip empty tables
        if not xds.coords and not xds.data_vars:
            continue
        if n in field_dict.keys():
            field_and_source_xds_dict[field_dict[n]] = xds
        else:
            sub_xds_dict[n] = xds

    return sub_xds_dict, field_and_source_xds_dict


def _get_data_name(xds, data_group):
    if "visibility" in xds.attrs["data_groups"][data_group]:
        data_name = xds.attrs["data_groups"][data_group]["visibility"]
    elif "spectrum" in xds.attrs["data_groups"][data_group]:
        data_name = xds.attrs["data_groups"][data_group]["spectrum"]
    else:
        error_message = (
            "No Visibility or Spectrum data variable found in data_group "
            + data_group
            + "."
        )
        logger.exception(error_message)
        raise ValueError(error_message)
    return data_name



from pathlib import Path
from typing import Dict, Tuple

import xarray as xr


from .subtables import subt_rename_ids
from ._tables.read import load_generic_table
from ._tables.read_subtables import read_delayed_pointing_table


def read_spw_ddi_ant_pol(inpath: str) -> Tuple[xr.Dataset]:
    """
    Reads the four metainfo subtables needed to load data chunks into xdss.

    Parameters
    ----------
    inpath : str
        MS path (main table)

    Returns
    -------
    Tuple[xr.Dataset]
        tuple with antenna, ddi, spw, and polarization setup subtables info
    """
    spw_xds = load_generic_table(
        inpath,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    )
    ddi_xds = load_generic_table(inpath, "DATA_DESCRIPTION")
    ant_xds = load_generic_table(
        inpath, "ANTENNA", rename_ids=subt_rename_ids["ANTENNA"]
    )
    pol_xds = load_generic_table(
        inpath, "POLARIZATION", rename_ids=subt_rename_ids["POLARIZATION"]
    )
    return ant_xds, ddi_xds, spw_xds, pol_xds


def finalize_chunks(
    infile: str, chunks: Dict[str, xr.Dataset], chunk_spec: Dict[str, slice]
) -> Dict[Tuple[int, int], xr.Dataset]:
    """
    Adds pointing variables to a dictionary of chunk xdss. This is
    intended to be added after reading chunks from an MS main table.

    Parameters
    ----------
    infile : str
        MS path (main table)
    chunks : Dict[str, xr.Dataset]
        chunk xdss
    chunk_spec : Dict[str, slice]
        specification of chunk to load

    Returns
    -------
    Dict[Tuple[int, int], xr.Dataset]
        dictionary of chunk xdss where every xds now has pointing
        data variables
    """
    pnt_name = "POINTING"
    pnt_path = Path(infile, pnt_name)
    if "time" in chunk_spec:
        time_slice = chunk_spec["time"]
    else:
        time_slice = None
    pnt_xds = read_delayed_pointing_table(
        str(pnt_path),
        rename_ids=subt_rename_ids.get(pnt_name, None),
        time_slice=time_slice,
    )

    if "time" not in pnt_xds.dims:
        return xr.Dataset()

    pnt_xds = pnt_xds.compute()

    pnt_chunks = {
        key: finalize_chunk_xds(infile, xds, pnt_xds)
        for _idx, (key, xds) in enumerate(chunks.items())
    }

    return pnt_chunks


def finalize_chunk_xds(
    infile: str, chunk_xds: xr.Dataset, pointing_xds: xr.Dataset
) -> xr.Dataset:
    """
    Adds pointing variables to one chunk xds.

    Parameters
    ----------
    infile : str
        MS path (main table)
    xds_chunk : xr.Dataset
        chunks xds
    pointing_xds : xr.Dataset
        pointing (sub)table xds

    Returns
    -------
    xr.Dataset
        chunk xds with pointing data variables interpolated form
        the pointing (sub)table
    """

    interp_pnt = pointing_xds.interp(time=chunk_xds.time, method="nearest")

    for var in interp_pnt.data_vars:
        chunk_xds[f"pointing_{var}"] = interp_pnt[var]

    return chunk_xds



from .processing_set import ProcessingSet
from .open_processing_set import open_processing_set
from .load_processing_set import load_processing_set, ProcessingSetIterator
from .convert_msv2_to_processing_set import (
    convert_msv2_to_processing_set,
    estimate_conversion_memory_and_cores,
)
from .measurement_set_xds import MeasurementSetXds

from .schema import SpectrumXds, VisibilityXds

__all__ = [
    "ProcessingSet",
    "MeasurementSetXds",
    "open_processing_set",
    "load_processing_set",
    "ProcessingSetIterator",
    "convert_msv2_to_processing_set",
    "estimate_conversion_memory_and_cores",
    "SpectrumXds",
    "VisibilityXds",
]



from ._zarr.xds_to_zarr import _write_zarr
from ._zarr.xds_from_zarr import _read_zarr
import numpy as np
import os
import xarray as xr
from ..._utils.zarr.common import _open_dataset


def _xds_to_zarr(xds: xr.Dataset, zarr_store: str):
    _write_zarr(xds, zarr_store)


def _xds_from_zarr(
    zarr_store: str, output: dict = {}, selection: dict = {}
) -> xr.Dataset:
    # supported key/values in output are:
    # "dv"
    #    what data variables should be returned as.
    #    "numpy": numpy arrays
    #    "dask": dask arrays
    # "coords"
    #    what coordinates should be returned as
    #    "numpy": numpy arrays
    return _read_zarr(zarr_store, output, selection)


def _load_image_from_zarr_no_dask(zarr_file: str, selection: dict) -> xr.Dataset:
    # At the moment image module does not support S3 file system. file_system=os is hardcoded.
    image_xds = _open_dataset(
        store=zarr_file, file_system=os, xds_isel=selection, load=True
    )
    for h in ["HISTORY", "_attrs_xds_history"]:
        history = os.sep.join([zarr_file, h])
        if os.path.isdir(history):
            image_xds.attrs["history"] = _open_dataset(history, load=True)
            break
    _iter_dict(image_xds.attrs)
    return image_xds


def _iter_dict(d: dict) -> None:
    for k, v in d.items():
        if isinstance(v, dict):
            keys = v.keys()
            if (
                len(keys) == 3
                and "_dtype" in keys
                and "_type" in keys
                and "_value" in keys
            ):
                d[k] = np.array(v["_value"], dtype=v["_dtype"])
            else:
                _iter_dict(v)
